# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .depot import MAX_FILE_SIZE
from .model import Object
from sqlalchemy import (
    ForeignKey, Column, String, DateTime, Text, Integer, LargeBinary)
import sqlalchemy.orm


class Protocol(Object):
    """ Das Protokoll. """

    title = Column(String(254))
    location = Column(String(254))
    attendees = Column(String(254))
    day = Column(DateTime)
    accounting_year = Column(Integer)
    details = sqlalchemy.orm.relation(
        'ProtocolDetail', uselist=True, backref='protocol',
        order_by="ProtocolDetail.id", cascade='all,delete')
    attachments = sqlalchemy.orm.relation(
        'ProtocolAttachment', uselist=True, backref='protocol',
        order_by="ProtocolAttachment.name", cascade='all,delete')
    commitments = sqlalchemy.orm.relation(
        'ProtocolCommitment', uselist=True, backref='protocol',
        cascade='all,delete')


class ProtocolDetail(Object):
    """ Ein Protokoll-Eintrag. """

    protocol_id = Column(
        Integer, ForeignKey('protocol.id'), nullable=False)
    duration = Column(Integer, nullable=False, server_default='0')
    message = Column(Text, nullable=False, server_default='')
    responsible = Column(String(2), nullable=False, server_default='')


class ProtocolAttachment(Object):
    """ Eine Anlage zum Protokoll. """

    protocol_id = Column(
        Integer, ForeignKey('protocol.id'), nullable=False)
    name = Column(String(100), default=u'')
    mimetype = Column(String(30), default=u'')
    size = Column(String(20), default=u'')
    data = Column(LargeBinary(MAX_FILE_SIZE))


class ProtocolCommitment(Object):
    """ Eine Anlage zum Protokoll. """

    protocol_id = Column(
        Integer, ForeignKey('protocol.id'), nullable=False)
    who = Column(String(254))
    what = Column(String(254))
    when = Column(String(254))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from ...conftest import import_members
import xlrd


def test_MemberAddView_1(browser, json_fixture):
    """It can add new members via JSON."""
    url = json_fixture.url()
    browser.login()
    browser.post('http://localhost{}'.format(url), data='')
    assert 'success' == browser.json['status']
    json_fixture.assertEqual(browser.json, 'data')


def test_mv_entrance_list_view(browser):
    import_members()
    browser.login()
    browser.open('http://localhost/members/mv_entrance_list')
    workbook = xlrd.open_workbook(file_contents=browser.contents)
    sheet = workbook.sheets()[0]
    assert sheet.cell(0, 0).value.endswith(' - Liste Einlass MV')
    assert 'Esser' == sheet.cell(2, 0).value
    assert 'Regina' == sheet.cell(2, 1).value
    assert '112' == sheet.cell(2, 2).value



"""Add accounting year for protocols.

Revision ID: 20588bf7b8a7
Revises: 55a653c1d5a0
Create Date: 2015-11-09 08:42:09.370420

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '20588bf7b8a7'
down_revision = '55a653c1d5a0'


def upgrade():
    op.add_column('protocol',
                  sa.Column('accounting_year', sa.Integer(), nullable=True))
    op.execute("""UPDATE protocol set accounting_year = 2014
                  WHERE day::text like '2014%';""")
    op.execute("""UPDATE protocol set accounting_year = 2015
                  WHERE day::text like '2015%';""")
    op.execute("""UPDATE protocol set accounting_year = 2016
                  WHERE day::text like '2016%';""")


def downgrade():
    op.drop_column('protocol', 'accounting_year')



# encoding=utf8
from __future__ import unicode_literals
from datetime import datetime
from sw.allotmentclub import User, Member, Organization
from sw.allotmentclub.model import ENGINE_NAME
import csv
import json
import os.path
import pkg_resources
import pytest
import risclog.sqlalchemy.testing
import transaction


class JSONFixture(object):

    basepath = ('sw.allotmentclub',
                '../../../spec/javascripts/fixtures/json/')
    _url = None

    def __init__(self, fixture, key):
        package, path = self.basepath
        self.fixture_path = pkg_resources.resource_filename(
            package, os.path.join(path, fixture))
        self.key = key
        self.fixture = self.load().get(self.key, {})

    def url(self, default=None):
        self._url = self.fixture.get('url', default)
        return self._url

    def load(self):
        with open(self.fixture_path, 'r') as fixture:
            return json.load(fixture)

    def write(self, content):
        with open(self.fixture_path, 'w') as fixture:
            json.dump(content, fixture, sort_keys=True, indent=4,
                      separators=(',', ': '))

    def assertEqual(self, data, data_key, save=False):
        if save:
            fixture = self.load()
            if self.key not in fixture:
                fixture[self.key] = {}
            fixture[self.key]['url'] = self._url
            fixture[self.key][data_key] = data.get(data_key, None)
            self.write(fixture)
            self.fixture = self.load()[self.key]
        expected = self.fixture[data_key] or []
        got = data.get(data_key, [])
        assert len(expected) == len(got)
        for item in expected:
            assert item in got

    def ajax(self, browser):
        """Simulate ajax call with data from fixture."""
        data = self.fixture['data']
        url = data['url'].replace('/api/', '')
        if data.get('type', data.get('method', '')).lower() == 'post':
            browser.post(
                'http://localhost{}'.format(url),
                data=data['data'],
                type=data['contentType'],
                xhr=True)
        else:
            browser.get('http://localhost{}'.format(url), xhr=True)


def assertFileEqual(generated_data, master_filename):
    from pdf_diff.command_line import compute_changes
    import pkg_resources
    import tempfile
    import os
    import pytest
    master_file = pkg_resources.resource_filename(
        'sw.allotmentclub.browser.tests', master_filename)
    handle, generated_file = tempfile.mkstemp(suffix='pdf')
    os.fdopen(handle, 'wb').write(generated_data)
    changes = compute_changes(master_file, generated_file)
    if changes:
        changes_text = ""
        for change in changes:
            if change == '*':
                changes_text += '\n\r'
                continue
            changes_text += '{}: {}\n\r'.format(
                change['pdf']['file'], change['text'])
        pytest.fail(
            'Generated pdf does not equal master: \n\r\n\r{}'.format(
                changes_text))
    else:
        os.remove(generated_file)


def pytest_configure(config):
    import sys
    sys._called_from_test = True


def pytest_unconfigure(config):
    import sys  # This was missing from the manual
    del sys._called_from_test


@pytest.fixture(scope='function')
def json_fixture(request):
    """Helper that retrieves JSON fixtures from files on disk."""
    return JSONFixture(
        request.module.__name__.split('.')[-1].replace('test_', '') + '.json',
        request.function.__name__.replace('test_', ''))


@pytest.fixture(scope='session')
def database_session(request):
    """Set up and tear down the sw.allotmentclub database.

    Returns the database utility object.
    """
    # create pytest fixture for database
    return risclog.sqlalchemy.testing.database_fixture_factory(
        request, 'sw.allotmentclub', ENGINE_NAME, create_all=True)


@pytest.fixture(scope='function')
def database(request, database_session):
    """Perform database setup and tear down for test function.

    Will empty all tables beforehand and close the session afterwards.

    Since this fixture is effectively used in every unit test, we also run the
    cleanup here.

    Returns the database utility object.
    """
    risclog.sqlalchemy.testing.database_test_livecycle_fixture_factory(request)
    return database_session


@pytest.fixture(scope='function')
def organization(request, database):
    """Fixture that creates an organization."""
    org = Organization.find_or_create(id=1, title='Leuna-Siedlung Roter See')
    database.session.flush()
    transaction.commit()
    return org


@pytest.fixture(scope='function')
def verwalter(request, organization, database):
    """Fixture creating a user with role Administrator."""
    user = User.find_or_create(
        username='admin', password='admin',
        vorname='Admin', nachname='istrator',
        unrestricted_access=True,
        organization_id=1)
    database.session.flush()
    transaction.commit()
    return user


@pytest.fixture(scope='function')
def user(request, organization, database):
    """Fixture creating a user with no role."""
    user = User.find_or_create(
        username='user', password='pass', nachname='Mittag',
        position='Vorsitzender', ort='Leuna', organization_id=1)
    database.session.flush()
    return user


@pytest.fixture(scope='function')
def member(request, database):
    """Fixture creating a member."""
    def delete_member():
        Member.query().filter(Member.lastname == 'Mittag').one().delete()
        database.session.flush()

    member = Member.create(firstname='Gerd', lastname='Mittag')
    database.session.flush()
    request.addfinalizer(delete_member)
    return member


class Amount(object):
    amount = None
    currency = None

    def __init__(self, amount, currency):
        self.amount = amount
        self.currency = currency


def import_bookings():
    from .account import import_transactions, BankingAccount
    account = BankingAccount.find_or_create(
        organization_id=1, number='3440000167')
    statements = []
    with open(pkg_resources.resource_filename(
            'sw.allotmentclub.tests', 'test_account_import.csv')) as f:
        reader = csv.reader(f, delimiter=';')
        for line in reader:
            (_, _, _, account_number, _, _, _, _, _, _, _, bic, _, iban, _, _,
             recipient, _, _, _, _, _, booking_date, _, value, currency, _, _,
             _, _, _, _, _, _, booking_text, _, _, purpose, _, _, _, _, _, _,
             _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
             _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _, _,
             _) = line
            statements.append({
                'date': datetime.strptime(booking_date, '%Y/%m/%d'),
                'currency': currency,
                'purpose': purpose,
                'amount': Amount(
                    float(value.replace('/1', ''))/10000,
                    currency),
                'applicant_name': recipient,
                'posting_text': booking_text,
                'applicant_iban': iban,
                'applicant_bin': bic,
                'customer_reference': None
            })
    import_transactions(statements, account)


def import_energy_meters():
    from sw.allotmentclub import Allotment, ElectricMeter, EnergyValue
    from sw.allotmentclub import EnergyPrice
    EnergyPrice.find_or_create(year=2014, value=1342354, bill=42312300,
                               price=3020, normal_fee=81700, power_fee=243300)
    EnergyPrice.find_or_create(year=2015, value=1354334, bill=42134200)
    for allot, number, v2013, v2014, power, disc, comment in [
            ['102', '318992603', '6893', '6893', '', '', ''],
            ['104', '136426011', '10019', '10634', 'X', '', ''],
            ['106', '21785640', '23154', '24207', '', '', ''],
            ['108', '81112116', '8165', '8411', '', '', ''],
            ['110', '31850195', '65811', '66345', '', '', ''],
            ['112', '20232757', '56221', '56371', '', '', ''],
            ['114', '364754', '7361', '7407', 'X', '', ''],
            ['118', '21292097', '935', '960', '', '', ''],
            ['122', '0063487695', '7028', '7988', 'X', '', ''],
            ['124', '4270447', '7671', '8033', '', '', ''],
            ['203', '21181284', '5933', '5997', '', '', ''],
            ['203', '3295328', '1307', '1349', '', '', 'Satanlage'],
            ['249', '20868068', '12115', '12115', '', '', ''],
            ['251', '20236014', '17339', '17352', '', '', 'Wasserpumpe'],
            ['328', '409120', '5075', '5075', '', '', ''],
            ['405', '8056675', '66018', '66098', '', '', '']]:
        allotment = Allotment.query().filter(
            Allotment.number == allot).one()
        meter = ElectricMeter.create(
            allotment=allotment,
            number=number,
            electric_power=bool(power),
            disconnected=bool(disc),
            comment=comment)
        for year, value in [(2013, v2013), (2014, v2014)]:
            value = EnergyValue.create(
                electric_meter=meter, year=year, value=int(value))
            value.update_member()
            value.update_usage()
            value.update_data()
            meter.energy_values.append(value)
        transaction.savepoint()


def import_members(max=99999):
    from sw.allotmentclub import Member, Allotment, Parcel
    for line in [
        ['60', '102', '', 'Frau', 'Ines', 'Groß', 'Mittelweg 7',
            '01458', 'Ottendorf-Okrilla', '', '', ''],
        ['62', '104', '', 'Herr', 'Reiner', 'Pfeil', 'Schillerstr. 42',
         '06247', 'Bad Lauchstädt', '', '034635 32731', ''],
        ['64', '106', '', 'Frau', 'Astrid', 'Ritter',
         'Brandenburgische Str.  27', '15366', 'Hönow', '', '', ''],
        ['67', '108', '', 'Herr', 'Sebastian', 'Wehrmann',
         'Geseniusstr. 34', '06110', 'Halle', '', '', '0172 1096832'],
        ['70', '110', '', 'Herr', 'Johannes', 'Hennig', 'A.-Einstein-'
         'Str. 15', '06237', 'Leuna', '', '03461 502682', ''],
        ['74', '112', '', 'Frau', 'Regina', 'Esser', 'Ringstr. 42',
         '06886', 'Wittenberg', '', '03491 662813', ''],
        ['76', '114', 'Dr.', 'Frau', 'Brigitte', 'Helbig', 'Wolfgang-'
         'Heinze-Str. 20', '04277', 'Leipzig', '', '0341 3520609', ''],
        ['83', '118', '', 'Frau', 'Margit ', 'Götze', 'Heinrich-Heine'
         '-Str. 19', '06237', 'Leuna', '', '03461 811244', ''],
        ['92/93', '122/ 225', '', 'Herr', 'Claus', 'Masthoff', 'Paul-'
         'Thiersch-Str. 16', '06124', 'Halle', '', '0345 6876407', ''],
        ['94/50', '124', '', 'Frau', 'Britta', 'Grimmling', 'Fors 190',
         '87391', 'Bollstabruk', 'Schweden', '', '0157 84943178'],
        ['150', '249/251', '', 'Herr', 'Lutz', 'Rösler',
         'Cloppenburger Str. 12', '06126', 'Halle', '', '', ''],
        ['100/137', '405/406', '', 'Frau', 'Marlies', 'Leutloff', 'Wei'
         'ßenfelser Str. 11c', '06231', 'Bad Dürrenberg', '', '', ''],
        ['', '', '', 'Herr', 'Günter', 'Tillack', 'Harry-S.-Truman-'
         'Allee 4', '14167', 'Berlin', '', '', '0162 9541236'],
        ['153', '328', '', '', '', 'Leuna Bungalowgemeinschaft Roter '
         'See e.V.', 'Postfach 4023', '06233', 'Leuna', '', '', ''],
        ['58', '203', '', '', '', 'Werkstatt', '', '', '', '', '', '']
    ]:
        if '/' in line[0] or '/' in line[1]:
            lines = [line[:], line[:]]
        else:
            lines = [line]

        if '/' in line[0]:
            p1, p2 = line[0].split('/')
            lines[0][0] = p1.strip()
            lines[1][0] = p2.strip()
        if '/' in line[1]:
            p1, p2 = line[1].split('/')
            lines[0][1] = p1.strip()
            lines[1][1] = p2.strip()

        for line in lines:
            member = Member.find_or_create(
                firstname=line[4], lastname=line[5])
            member.title = line[2]
            member.appellation = line[3]
            member.street = line[6]
            member.zip = line[7]
            member.city = line[8]
            member.country = line[9] or 'Deutschland'
            member.phone = line[10]
            member.mobile = line[11]

            if line[1].strip():
                allotment = Allotment.find_or_create(
                    number=int(line[1]), member=member)
                if line[0].strip():
                    Parcel.find_or_create(number=int(line[0]),
                                          allotment=allotment)
            transaction.savepoint()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .log import log_with_user, auth_log
from .model import Object, Organization
from datetime import date, timedelta
from fints.client import FinTS3PinTanClient
from sqlalchemy import Column, Integer, String, ForeignKey, DateTime, Boolean
from sqlalchemy import Date
import kontocheck
import pyramid.threadlocal
import sqlalchemy.orm
import transaction


class BankingAccount(Object):
    """ Ein Bankkonto. """

    number = Column(String(10))
    name = Column(String(30))
    last_import = Column(DateTime)


class BookingKind(Object):
    """ Ein Buchungstyp. """

    title = Column(String(50), nullable=False)
    shorttitle = Column(String(10), nullable=True)
    banking_account_id = Column(
        Integer, ForeignKey('bankingaccount.id'), nullable=True)
    banking_account = sqlalchemy.orm.relation(
        'BankingAccount', uselist=False, backref='booking_kinds')


class Booking(Object):
    """ Eine Buchung. """

    banking_account_id = Column(
        Integer, ForeignKey('bankingaccount.id'), nullable=False)
    banking_account = sqlalchemy.orm.relation(
        'BankingAccount', uselist=False, backref='bookings')

    booking_day = Column(Date)
    booking_text = Column(String)
    purpose = Column(String)
    recipient = Column(String)
    iban = Column(String(34))
    bic = Column(String(11))
    value = Column(Integer)

    is_splitted = Column(Boolean, default=False, nullable=False)
    splitted_from_id = Column(Integer, ForeignKey('booking.id'), nullable=True)
    accounting_year = Column(Integer)
    ignore_in_reporting = Column(Boolean, default=False, nullable=False)

    member_id = Column(Integer, ForeignKey('member.id'), nullable=True)
    member = sqlalchemy.orm.relation(
        'Member', uselist=False, backref='bookings')
    kind_id = Column(Integer, ForeignKey('bookingkind.id'), nullable=True)
    kind = sqlalchemy.orm.relation('BookingKind', uselist=False)

    def split(self, new_value):
        value = self.value if self.value is not None else 0
        if value > 0 and new_value < 0:
            return 'Neuer Wert muss positiv sein.'
        if value < 0 and new_value > 0:
            return 'Neuer Wert muss negativ sein.'
        if value > 0 and new_value >= self.value:
            return 'Neuer Wert ist zu groß.'
        if value < 0 and new_value <= self.value:
            return 'Neuer Wert ist zu klein.'
        if new_value == 0:
            return 'Neuer Wert muss größer oder kleiner 0 sein.'
        result = []
        for value in [new_value, self.value - new_value]:
            result.append(Booking.create(
                banking_account=self.banking_account,
                booking_day=self.booking_day,
                booking_text=self.booking_text,
                purpose=self.purpose,
                recipient=self.recipient,
                iban=self.iban,
                bic=self.bic,
                accounting_year=self.accounting_year,
                value=value,
                splitted_from_id=self.id,
                member=self.member,
                kind=self.kind))
        self.is_splitted = True
        return result


class SEPASammler(Object):
    """Ein SEPASammler."""

    pmtinfid = Column(String(100))
    booking_day = Column(Date)
    accounting_year = Column(Integer)
    kind_id = Column(Integer, ForeignKey('bookingkind.id'), nullable=True)
    kind = sqlalchemy.orm.relation('BookingKind', uselist=False)


class SEPASammlerEntry(Object):
    """Defines which members are active in a SEPA Sammler."""

    sepa_sammler_id = Column(
        Integer, ForeignKey('sepasammler.id'), nullable=True)
    sepasammler = sqlalchemy.orm.relation(
        'SEPASammler', uselist=False, backref='entries')
    ignore_in_reporting = Column(Boolean, default=False, nullable=False)
    member_id = Column(Integer, ForeignKey('member.id'), nullable=True)
    member = sqlalchemy.orm.relation('Member', uselist=False)
    value = Column(Integer)


class GrundsteuerB(Object):

    parcel_id = Column(
        Integer, ForeignKey('parcel.id'), nullable=False)
    parcel = sqlalchemy.orm.relation('Parcel', uselist=False)
    value = Column(Integer)


class Abwasser(Object):

    parcel_id = Column(
        Integer, ForeignKey('parcel.id'), nullable=False)
    parcel = sqlalchemy.orm.relation('Parcel', uselist=False)
    value = Column(Integer)


def create_fints_client():
    settings = pyramid.threadlocal.get_current_registry().settings
    return FinTS3PinTanClient(
        settings.get('banking.blz'),
        settings.get('banking.kto'),
        settings.get('banking.pin'),
        settings.get('banking.url'))


def import_transactions_from_fints(user=None):
    kontocheck.lut_load(9)
    client = create_fints_client()
    accounts = client.get_sepa_accounts()
    for account in accounts:
        try:
            local_account = (
                BankingAccount.query()
                .filter(BankingAccount.number == account.accountnumber)
                .one())
        except sqlalchemy.orm.exc.NoResultFound:
            continue
        statements = client.get_statement(
            account, date.today() - timedelta(days=15), date.today())
        import_transactions(
            [stmt.data for stmt in statements], local_account, user)


def import_transactions(statements, account, user=None):
    count = 0
    for id, statement in enumerate(statements):
        try:
            result = add_transaction(statement, account)
        except Exception as e:
            log_with_user(
                auth_log.info, user,
                'hat Bank-Import ausgeführt: Fehler ({}).'.format(e))
            transaction.commit()
            raise
        if isinstance(result, int):
            count += result
    if count and user:
        log_with_user(
            auth_log.info, user,
            'hat Bank-Import ausgeführt: {} neue Buchungen.'.format(count))
    transaction.commit()


def get_sepa_sammlers(data, account):
    if data['posting_text'] != 'SAMMEL-LS-EINZUG':
        return
    pmtinfid = data['customer_reference'].splitlines()[1]
    try:
        sepasammler = (
            SEPASammler.query()
            .filter(SEPASammler.pmtinfid == pmtinfid)
            .filter(SEPASammler.organization_id == account.organization_id)
            .one())
    except Exception:
        raise ValueError("Keine Sammler zu PmtInfId '{}' gefunden.".format(
            pmtinfid))
    sammlers = (
        SEPASammlerEntry.query()
        .filter(SEPASammlerEntry.sepasammler == sepasammler)
        .filter(
            SEPASammlerEntry.organization_id == account.organization_id)
        .all())
    if not sammlers:
        raise ValueError(
            "Keine Sammler-Einträge zu PmtInfId '{}' gefunden.".format(
                pmtinfid))
    return sammlers


def add_transaction(data, account):
    from sw.allotmentclub import Member
    accounting_year = data['date'].year
    org_title = None
    org = Organization.get(account.organization_id)
    if org:
        org_title = org.title
    if data['currency'] != 'EUR':
        raise ValueError(
            "Couldn't import booking {} because its not in Eur.".format(
                data['purpose']))
    value = int(data['amount'].amount * 10000)
    sammlers = get_sepa_sammlers(data, account)
    booking = Booking.find_or_create(
        organization_id=account.organization_id,
        banking_account=account,
        booking_day=data['date'],
        purpose=data['purpose'],
        recipient=(
            data['applicant_name']
            if data['applicant_name'] is not None else org_title),
        value=value)
    if booking.id:
        return
    booking.accounting_year = accounting_year
    booking.booking_text = data['posting_text']
    booking.iban = data['applicant_iban']
    booking.bic = data['applicant_bin']
    if booking.iban:
        try:
            booking.member = (
                Member.query()
                .filter(Member.iban == booking.iban)
                .filter(Member.organization_id == booking.organization_id)
                .one())
        except Exception:
            pass
    if booking.member:
        try:
            debit_booking = (
                Booking.query()
                .filter(Booking.value == 0 - booking.value)
                .filter(Booking.member == booking.member)
                .filter(Booking.accounting_year == booking.accounting_year)
                .filter(Booking.organization_id == booking.organization_id)
                .one())
        except Exception:
            pass
        else:
            booking.kind = debit_booking.kind
    if booking.booking_text == 'SAMMEL-LS-EINZUG':
        booking.recipient = org_title
        sum_ = 0
        for sammler in sammlers:
            sum_ += sammler.value
        if (booking.value - 100) > sum_:
            imported, open_ = booking.split(sum_)
            imported.booking_text = imported.booking_text + ' Verrechnet'
            open_.organization_id = booking.organization_id
            imported.organization_id = booking.organization_id
        sum_ += sammler.value
    return 1



"""Add signature for hs.

Revision ID: 96017ac786bf
Revises: 769e10087d9e
Create Date: 2017-07-28 13:21:10.263997

"""
from alembic import op
import base64
import pkg_resources

# revision identifiers, used by Alembic.
revision = '96017ac786bf'
down_revision = '769e10087d9e'


def upgrade():
    signature = pkg_resources.resource_stream(
        'sw.allotmentclub.signatures', 'hs.png').read()
    signature = 'data:application/png;base64,{}'.format(
        base64.b64encode(signature).decode('utf-8'))
    op.execute(u"""UPDATE public.user SET
                  signature = '{signature}'
                  WHERE username = 'hs';""".format(signature=signature))


def downgrade():
    op.execute(u"""UPDATE public.user SET
                  signature = ''
                  WHERE username = 'hs';""")



"""Add SEPASammler grouping items

Revision ID: b7de0b4c2bb6
Revises: 7136a99f680d
Create Date: 2016-06-28 08:47:19.610508

"""

from alembic import op
from sqlalchemy.schema import Sequence, CreateSequence
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'b7de0b4c2bb6'
down_revision = '7136a99f680d'


def upgrade():
    op.execute("""ALTER TABLE sepasammler RENAME TO sepasammlerentry;""")
    op.execute("CREATE TABLE sepasammler AS SELECT * FROM sepasammlerentry;")
    op.execute("""DELETE FROM sepasammler""")
    op.execute("""ALTER SEQUENCE sepasammler_id_seq RENAME TO
               sepasammlerentry_id_seq""")
    op.execute("""ALTER INDEX sepasammler_pkey RENAME TO
               sepasammlerentry_pkey""")
    op.execute(CreateSequence(Sequence('sepasammler_id_seq')))
    op.alter_column(
        "sepasammler", "id", nullable=False,
        server_default=sa.text("nextval('sepasammler_id_seq'::regclass)"))
    op.create_unique_constraint(u'sepasammler_pkey', 'sepasammler', ['id'])
    op.drop_column(u'sepasammler', 'ignore_in_reporting')
    op.drop_column(u'sepasammler', 'value')
    op.drop_column(u'sepasammler', 'member_id')
    op.execute("""INSERT into sepasammler (pmtinfid, booking_day,
               accounting_year, kind_id) (SELECT distinct pmtinfid,
               booking_day, accounting_year, kind_id FROM sepasammlerentry)""")
    op.create_foreign_key(
        None, 'sepasammler', 'bookingkind', ['kind_id'], ['id'])
    op.add_column(
        'sepasammlerentry',
        sa.Column('sepa_sammler_id', sa.Integer(), nullable=True))
    op.execute("""UPDATE sepasammlerentry SET sepa_sammler_id = (
        SELECT id FROM sepasammler WHERE pmtinfid = sepasammlerentry.pmtinfid
    )""")
    op.create_foreign_key(
        None, 'sepasammlerentry', 'sepasammler', ['sepa_sammler_id'], ['id'])
    op.drop_column('sepasammlerentry', 'kind_id')
    op.drop_column('sepasammlerentry', 'accounting_year')
    op.drop_column('sepasammlerentry', 'pmtinfid')
    op.drop_column('sepasammlerentry', 'booking_day')


def downgrade():
    raise NotImplementedError("Not possible.")



# encoding=utf8
"""Add payer name to direct debit.

Revision ID: cfc9142d7e69
Revises: c3c978348cbf
Create Date: 2016-02-15 07:52:19.049448

"""
from __future__ import unicode_literals
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'cfc9142d7e69'
down_revision = 'c3c978348cbf'


def upgrade():
    op.add_column(
        'member',
        sa.Column('direct_debit_account_holder',
                  sa.String(length=100), nullable=True))
    op.execute(
        """UPDATE member SET
            firstname = 'Thomas',
            lastname = 'Sonntag',
            street = 'Auenwinkel 2',
            zip = '06237',
            city = 'Leuna/OT Maßlau',
            email = 'tschloeffel@freenet.de',
            direct_debit_account_holder = 'Schlöffel, Heinz'
     WHERE id = 109;""")


def downgrade():
    op.execute(
        """UPDATE member SET
            firstname = 'Heinz',
            lastname = 'Schlöffel',
            street = 'Auenwinkel 7',
            zip = '06237',
            city = 'Leuna',
            email = 'heinz.schloeffel@freenet.de'
     WHERE id = 109;""")
    op.drop_column('member', 'direct_debit_account_holder')



# encoding=utf-8
from __future__ import unicode_literals
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.view import view_config
from sw.allotmentclub.browser.base import View, route_permitted, route_url


NAVIGATION_ITEMS = (
    # Title and route
    ('Startseite', 'home', 'fa-home', None),
    ('Ablage', 'depots', 'fa-inbox', None),
    ('Aushänge', 'bulletins', 'fa-thumb-tack', None),
    ('Protokolle', 'protocols', 'fa-book', None),
    ('Schlüsselbuch', 'keylists', 'fa-key', None),
    ('Arbeitseinsätze', 'assignments', 'fa-wrench', (
        ('Übersicht', 'assignments'),
        ('Abrechnung', 'member_assignments'),
    )),
    ('Mitglieder', 'member_list', 'fa-user', (
        ('Alle', 'member_list'),
        ('Pächter', 'member_list_leased'),
        ('Trinkwasser', 'member_list_tap_water'),
        ('Lageplan', 'map'),
    )),
    ('Postversand', 'mail_list_inbox', 'fa-envelope-square', (
        ('Posteingang', 'mail_list_inbox'),
        ('Entwürfe', 'mail_list_drafts'),
        ('Gesendet', 'mail_list_sent'),
        ('Externe Empfänger', 'externals'),
    )),
    ('Energie', 'electricity_list', 'fa-lightbulb-o', (
        ('Stromzähler', 'electricity_list'),
        ('Abrechnungen', 'global_energy_value_list'),
        ('Kennzahlen', 'energy_price'),
    )),
    ('Abwasser', 'waste_water', 'fa-recycle', None),
    ('Grundsteuer B', 'property_tax_b', 'fa-university', None),
    ('Finanzen', 'booking_list', 'fa-credit-card', (
        ('Konto-Typen', 'banking_account_list'),
        ('Vereins-Konto', 'booking_list'),
        ('Mitgliedskonten', 'member_account_list'),
        ('SEPA-Lastschriften', 'sepa_direct_debit'),
        ('SEPA-Sammler', 'sepa_sammler_list'),
    )),
    ('Berechtigungen', 'access_authority', 'fa-unlock', None),
)


@view_config(route_name='navigation',
             permission=NO_PERMISSION_REQUIRED,
             renderer='json')
class NavigationView(View):

    def __call__(self):
        result = []
        for title, route, css, sublinks in NAVIGATION_ITEMS:
            if not route_permitted(route, self.request):
                continue
            url = route_url(route, self.request)
            entry = {'route': route,
                     'title': title,
                     'url': url,
                     'css': css}
            if sublinks is not None:
                subentries = []
                for title, route in sublinks:
                    if not route_permitted(route, self.request):
                        continue
                    subentries.append(
                        {'route': route,
                         'title': title,
                         'url': route_url(route, self.request)})
                if subentries:
                    entry['subs'] = subentries
            result.append(entry)
        return {'status': 'success', 'data': result}



# encoding=utf-8
from __future__ import unicode_literals
from ..log import app_log
from .base import string_agg, date
from io import BytesIO
from pyramid.response import Response
from pyramid.view import view_config
from sw.allotmentclub import Parcel, Allotment, Member
from svglib.svglib import svg2rlg
from reportlab.graphics import renderPDF
from reportlab.pdfgen.canvas import Canvas
from datetime import datetime
import pkg_resources
import sw.allotmentclub.browser.base
import tempfile
import os


class Query(sw.allotmentclub.browser.base.Query):

    def select(self):
        return (
            self.db.query(
                Parcel.number.label('Flurstück'),
                string_agg(Allotment.number).label('Bungalow'),
                Member.lastname.label('Nachname'),
                Member.firstname.label('Vorname'),
            )
            .select_from(Parcel)
            .join(Allotment)
            .join(Member)
            .group_by(Parcel.number, Member.firstname, Member.lastname))


@view_config(route_name='map', renderer='json', permission='view')
class MapView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    available_actions = [
        dict(url='map_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]

    def __call__(self):
        result = super(MapView, self).__call__()
        map_data = {}
        for parcel, allotment, ln, fn in sw.allotmentclub.json_result(
                result['data']['data']):
            if fn:
                ln += ', {}.'.format(fn[0])
            map_data[str(parcel)] = [str(allotment), ln]
        result['data']['map_data'] = map_data
        with open(pkg_resources.resource_filename(
                'sw.allotmentclub.browser', 'lageplan.svg')) as map_:
            result['data']['map'] = map_.read()
        return result


@view_config(route_name='map_download', permission='view')
class MapDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        data = self.request.POST['svg']
        msg = 'Lageplan vom %s' % date(datetime.now())
        svg_handle, svg_filename = tempfile.mkstemp(suffix='svg')
        app_log.info('Map written to `{}`.'.format(svg_filename))
        try:
            os.fdopen(svg_handle, 'wb').write(data.encode('utf-8'))
            d = renderPDF.renderScaledDrawing(svg2rlg(svg_filename))
            output = BytesIO()
            c = Canvas(output)
            c.setTitle(msg)
            c.setPageSize((d.width, d.height))
            renderPDF.draw(d, c, 0, 0)
            c.showPage()
            c.save()
        finally:
            os.remove(svg_filename)
        output.seek(0)
        data = output.read()
        return Response(
            content_type='application/pdf',
            content_disposition='attachment; filename="Lageplan.pdf"',
            content_length=len(data),
            charset=b'utf-8',
            body=data)



"""Delete due payments for members that have sold their allotment.

Revision ID: 45d928fd0cb7
Revises: cfc9142d7e69
Create Date: 2016-03-14 09:29:49.734108

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '45d928fd0cb7'
down_revision = u'cfc9142d7e69'


def upgrade():
    op.execute("""DELETE FROM booking
                  WHERE accounting_year = 2016 and member_id in (132, 128);""")
    op.execute("""UPDATE member SET bic = 'DRESDEFF800' WHERE id = 118;""")


def downgrade():
    pass



# encoding=utf8
"""Remove year from BookingKind.

Revision ID: 32b7b30817d6
Revises: 1fb65d65969
Create Date: 2015-11-18 07:30:07.933913

"""

# revision identifiers, used by Alembic.
from __future__ import unicode_literals
from alembic import op
import sqlalchemy as sa

revision = '32b7b30817d6'
down_revision = '1fb65d65969'


def upgrade():
    op.execute("""DELETE FROM bookingkind WHERE year = 2016;""")
    op.add_column(
        'bookingkind',
        sa.Column('shorttitle', sa.String(length=10), nullable=True))
    op.alter_column('bookingkind', 'title', type_=sa.String(length=50))
    op.drop_column('bookingkind', 'year')
    op.drop_column('bookingkind', 'value_per_member')
    op.execute("""UPDATE bookingkind SET shorttitle = 'NGA'
                  WHERE title = 'Fehl.Arb.Std';""")
    op.execute("""UPDATE bookingkind SET title = 'Nicht geleistete
      Arbeitsstunden' WHERE shorttitle = 'NGA';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'MGB'
                  WHERE title = 'Mitgliedsbeitrag';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'EA1'
                  WHERE title = 'Energieabschlag I';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'EA2'
                  WHERE title = 'Energieabschlag II';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'EAR'
                  WHERE title = 'Energieabrechnung';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'GSB'
                  WHERE title = 'Grundsteuer B';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'ABW'
                  WHERE title = 'Abwasser';""")
    op.execute("""UPDATE bookingkind SET shorttitle = 'AUF'
                  WHERE title = 'Aufnahmegebühr';""")
    op.execute("""UPDATE booking SET kind_id = 1 WHERE id = 953;""")
    op.execute("""UPDATE booking SET kind_id = 2 WHERE id = 954;""")


def downgrade():
    op.add_column(
        'bookingkind',
        sa.Column('year', sa.INTEGER(), nullable=True))
    op.add_column(
        'bookingkind',
        sa.Column('value_per_member', sa.INTEGER(), nullable=True))
    op.drop_column('bookingkind', 'shorttitle')



"""Allow selecting multiple members on messages.

Revision ID: d6e0f091a932
Revises: 5b929a47d6ae
Create Date: 2016-04-21 08:49:21.445319

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd6e0f091a932'
down_revision = '5b929a47d6ae'


def upgrade():
    op.create_table(
        'members_messages',
        sa.Column('member_id', sa.Integer(), nullable=True),
        sa.Column('message_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['member_id'], ['member.id'], ),
        sa.ForeignKeyConstraint(['message_id'], ['message.id'], )
    )
    op.execute(
        """INSERT into members_messages (message_id, member_id)
             (SELECT id, member_id FROM message WHERE member_id is not null);
        """)
    op.drop_constraint('message_member_id_fkey', 'message', type_='foreignkey')
    op.drop_column('message', 'member_id')


def downgrade():
    op.add_column(
        'message',
        sa.Column('member_id', sa.INTEGER(), autoincrement=False,
                  nullable=True))
    op.create_foreign_key(
        'message_member_id_fkey', 'message', 'member', ['member_id'], ['id'])
    op.execute("""UPDATE message SET member_id = (
                      SELECT member_id FROM members_messages
                          WHERE message_id = message.id);""")
    op.drop_table('members_messages')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import base64
import pytest
import transaction

MINIMAL_TEST_PDF_CONTENTS = (
    "JVBERi0xLjEKJcKlwrHDqwoKMSAwIG9iagogIDw8IC9UeXBlIC9DYXRhbG9nCiAgICAgL1Bh"
    "Z2VzIDIgMCBSCiAgPj4KZW5kb2JqCgoyIDAgb2JqCiAgPDwgL1R5cGUgL1BhZ2VzCiAgICAg"
    "L0tpZHMgWzMgMCBSXQogICAgIC9Db3VudCAxCiAgICAgL01lZGlhQm94IFswIDAgMzAwIDE0"
    "NF0KICA+PgplbmRvYmoKCjMgMCBvYmoKICA8PCAgL1R5cGUgL1BhZ2UKICAgICAgL1BhcmVu"
    "dCAyIDAgUgogICAgICAvUmVzb3VyY2VzCiAgICAgICA8PCAvRm9udAogICAgICAgICAgIDw8"
    "IC9GMQogICAgICAgICAgICAgICA8PCAvVHlwZSAvRm9udAogICAgICAgICAgICAgICAgICAv"
    "U3VidHlwZSAvVHlwZTEKICAgICAgICAgICAgICAgICAgL0Jhc2VGb250IC9UaW1lcy1Sb21h"
    "bgogICAgICAgICAgICAgICA+PgogICAgICAgICAgID4+CiAgICAgICA+PgogICAgICAvQ29u"
    "dGVudHMgNCAwIFIKICA+PgplbmRvYmoKCjQgMCBvYmoKICA8PCAvTGVuZ3RoIDU1ID4+CnN0"
    "cmVhbQogIEJUCiAgICAvRjEgMTggVGYKICAgIDAgMCBUZAogICAgKEhlbGxvIFdvcmxkKSBU"
    "agogIEVUCmVuZHN0cmVhbQplbmRvYmoKCnhyZWYKMCA1CjAwMDAwMDAwMDAgNjU1MzUgZiAK"
    "MDAwMDAwMDAxOCAwMDAwMCBuIAowMDAwMDAwMDc3IDAwMDAwIG4gCjAwMDAwMDAxNzggMDAw"
    "MDAgbiAKMDAwMDAwMDQ1NyAwMDAwMCBuIAp0cmFpbGVyCiAgPDwgIC9Sb290IDEgMCBSCiAg"
    "ICAgIC9TaXplIDUKICA+PgpzdGFydHhyZWYKNTY1CiUlRU9GCg==")


@pytest.fixture(scope='function')
def setUp():
    from sw.allotmentclub import Member, Message, User, SentMessageInfo
    mustermann = Member.create(lastname="Mustermann", firstname="Max",
                               email='max@mustermann.de')
    user = User.create(username='sw')
    msg = Message.create(id=242, members=[mustermann], user=user,
                         accounting_year=2016, subject="Test", body="")
    SentMessageInfo.create(message=msg, tag='foo@vorstand.roter-see.de',
                           address='max@mustermann.de')
    transaction.commit()


def test__mail__postmark_webhook_1(browser, setUp):
    """It saves the open status when messages opens."""
    from sw.allotmentclub import SentMessageInfo
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"FirstOpen": True,
                       "Client": {"Name": "Chrome 35.0.1916.153",
                                  "Company": "Google",
                                  "Family": "Chrome"},
                       "OS": {"Name": "OS X 10.7 Lion",
                              "Company": "Apple Computer, Inc.",
                              "Family": "OS X 10"},
                       "Platform": "WebMail",
                       "UserAgent": "Mozilla\/5.0 (Macintosh Safari\/537.36)",
                       "ReadSeconds": 5,
                       "Geo": {"CountryISOCode": "RS",
                               "Country": "Serbia",
                               "RegionISOCode": "VO",
                               "Region": "Autonomna Pokrajina Vojvodina",
                               "City": "Novi Sad",
                               "Zip": "21000",
                               "Coords": "45.2517,19.8369",
                               "IP": "188.2.95.4"},
                       "MessageID": "asdadasd",
                       "ReceivedAt": "2016-11-23T07:03:45",
                       "Tag": "foo@vorstand.roter-see.de",
                       "Recipient": "max@mustermann.de"})
    status = SentMessageInfo.query().one().status
    assert (
        'Geöffnet am 23.11.2016 13:03 Uhr in Novi Sad (IP: 188.2.95.4) '
        'für 5 Sekunden.' == status)


def test__mail__postmark_webhook_2(browser, setUp):
    """It creates new status for addresses not in db."""
    from sw.allotmentclub import SentMessageInfo
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"ReadSeconds": 5,
                       "Geo": {"City": "Novi Sad",
                               "IP": "188.2.95.4"},
                       "Tag": "foo@vorstand.roter-see.de",
                       "ReceivedAt": "2014-06-01T12:00:00",
                       "Recipient": "sebastian.wehrmann@icloud.com"})
    assert 2 == len(SentMessageInfo.query().all())
    status = SentMessageInfo.query().filter(
        SentMessageInfo.address == 'sebastian.wehrmann@icloud.com').one()
    assert status is not None


def test__mail__postmark_webhook_3(browser, setUp):
    """It creates bounce status when message was bounced."""
    from sw.allotmentclub import SentMessageInfo
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"ID": 42,
                       "Type": "HardBounce",
                       "TypeCode": 1,
                       "Name": "Hard bounce",
                       "Tag": "foo@vorstand.roter-see.de",
                       "MessageID": "883953f4-6105-42a2-a16a-77a8eac79483",
                       "ServerId": 23,
                       "Description": (
                           "The server was unable to deliver your "
                           "message (ex: unknown user, mailbox not found)."),
                       "Details": "Test bounce details",
                       "Email": "max@mustermann.de",
                       "From": "vorstand@roter-see.de",
                       "BouncedAt": "2014-08-01T13:28:10.2735393-04:00",
                       "DumpAvailable": True,
                       "Inactive": True,
                       "CanActivate": True,
                       "Subject": "Test subject"})
    status = SentMessageInfo.query().one().status
    assert ('Hard bounce am 01.08.2014 19:28 Uhr: The server was unable to '
            'deliver your message (ex: unknown user, mailbox not '
            'found).' == status)


def test__mail__postmark_webhook_4(browser, setUp):
    """It creates delivered status when message was delivered."""
    from sw.allotmentclub import SentMessageInfo
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"ServerId": 23,
                       "MessageID": "883953f4-6105-42a2-a16a-77a8eac79483",
                       "Recipient": "max@mustermann.de",
                       "Tag": "foo@vorstand.roter-see.de",
                       "DeliveredAt": "2014-08-01T13:28:10.2735393-04:00",
                       "Details": "Test delivery webhook details"})
    status = SentMessageInfo.query().one().status
    assert 'Zugestellt am 01.08.2014 19:28 Uhr.' == status


def test__mail__postmark_webhook_5(browser, setUp):
    """It updates status for multiple requests."""
    from sw.allotmentclub import SentMessageInfo
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"Recipient": "max@mustermann.de",
                       "Tag": "foo@vorstand.roter-see.de",
                       "DeliveredAt": "2014-08-01T13:28:10.2735393-04:00"})
    browser.post('http://localhost/mail/postmark_open_tracking_webhook',
                 data={"ReadSeconds": 5,
                       "Geo": {"City": "Novi Sad",
                               "IP": "188.2.95.4"},
                       "ReceivedAt": "2016-11-23T07:03:45",
                       "Tag": "foo@vorstand.roter-see.de",
                       "Recipient": "max@mustermann.de"})
    status = SentMessageInfo.query().one().status
    assert (
        'Geöffnet am 23.11.2016 13:03 Uhr in Novi Sad (IP: 188.2.95.4) '
        'für 5 Sekunden.' == status)


def test__mail__postmark_inbound_webhook_1(browser):
    """It creates an inbound email for every request."""
    from sw.allotmentclub import Message, User
    user = User.find_or_create(username='system')
    browser.post(
        'http://localhost/mail/postmark_inbound_webhook',
        data={"FromName": "Postmarkapp Support",
              "From": "support@postmarkapp.com",
              "FromFull": {
                  "Email": "support@postmarkapp.com",
                  "Name": "Postmarkapp Support",
                  "MailboxHash": ""
              },
              "To": ("\"Firstname Lastname\" "
                     "<yourhash+SampleHash@inbound.postmarkapp.com>"),
              "ToFull": [
                  {
                      "Email": "yourhash+SampleHash@inbound.postmarkapp.com",
                      "Name": "Firstname Lastname",
                      "MailboxHash": "SampleHash"
                  }
              ],
              "Cc": ("\"First Cc\" "
                     "<firstcc@postmarkapp.com>, secondCc@postmarkapp.com>"),
              "CcFull": [
                  {
                      "Email": "firstcc@postmarkapp.com",
                      "Name": "First Cc",
                      "MailboxHash": ""
                  },
                  {
                      "Email": "secondCc@postmarkapp.com",
                      "Name": "",
                      "MailboxHash": ""
                  }
              ],
              "OriginalRecipient": "SampleHash@inbound.postmarkapp.com",
              "Subject": "Test subject",
              "MessageID": "73e6d360-66eb-11e1-8e72-a8904824019b",
              "ReplyTo": "replyto@postmarkapp.com",
              "MailboxHash": "SampleHash",
              # "Date": "Fri, 1 Aug 2014 16:45:32 -04:00",
              "Date": "Sat, 19 Aug 2017 19:37:23 +0200 (Mitteleurop?ische",
              "TextBody": "This is a test text body.",
              "HtmlBody": ("<html><body><p>"
                           "This is a test html body."
                           "<\/p><\/body><\/html>"),
              "StrippedTextReply": "This is the reply text",
              "Tag": "TestTag",
              "Headers": [
                  {
                      "Name": "X-Header-Test",
                      "Value": ""
                  },
                  {
                      "Name": "X-Spam-Status",
                      "Value": "No"
                  },
                  {
                      "Name": "X-Spam-Score",
                      "Value": "-0.1"
                  },
                  {
                      "Name": "X-Spam-Tests",
                      "Value": "DKIM_SIGNED,DKIM_VALID,DKIM_VALID_AU,SPF_PASS"
                  }
              ],
              "Attachments": [
                  {
                      "Name": "test.pdf",
                      "Content": MINIMAL_TEST_PDF_CONTENTS,
                      "ContentType": "application/pdf",
                      "ContentLength": 739
                  }
              ]})
    mail = Message.query().one()
    assert 'support@postmarkapp.com' == mail.externals[0].email
    assert 'Postmarkapp Support' == mail.externals[0].lastname
    assert [] == mail.members
    assert "Test subject" == mail.subject
    assert user == mail.user
    assert mail.inbound is True
    assert ("<html><body><p>"
            "This is a test html body."
            "<\/p><\/body><\/html>" == mail.body)
    assert 1 == len(mail.attachments)
    assert MINIMAL_TEST_PDF_CONTENTS.encode() == base64.b64encode(
        mail.attachments[0].data)
    assert 'application/pdf' == mail.attachments[0].mimetype
    assert 'test.pdf' == mail.attachments[0].filename
    assert int(mail.attachments[0].size) == len(mail.attachments[0].data)



"""Add external recipient for messages.

Revision ID: 4841df08d22b
Revises: 45d928fd0cb7
Create Date: 2016-03-30 12:44:52.036678

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4841df08d22b'
down_revision = '45d928fd0cb7'


def upgrade():
    op.create_table(
        'externalrecipient',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('title', sa.String(length=20), nullable=True),
        sa.Column('appellation', sa.String(length=20), nullable=True),
        sa.Column('firstname', sa.String(length=50), nullable=True),
        sa.Column('lastname', sa.String(length=50), nullable=True),
        sa.Column('street', sa.String(length=100), nullable=True),
        sa.Column('zip', sa.String(length=6), nullable=True),
        sa.Column('city', sa.String(length=50), nullable=True),
        sa.Column('country', sa.String(length=50), nullable=True),
        sa.Column('email', sa.String(length=100), nullable=True),
        sa.PrimaryKeyConstraint('id')
    )
    op.add_column(
        u'message', sa.Column('external_id', sa.Integer(), nullable=True))
    op.create_foreign_key(
        'externalrecipient_message',
        'message', 'externalrecipient', ['external_id'], ['id'])


def downgrade():
    op.drop_constraint(
        'externalrecipient_message', 'message', type_='foreignkey')
    op.drop_column(u'message', 'external_id')
    op.drop_table('externalrecipient')



"""Some missing changes to database.

Revision ID: 88942e83a048
Revises: b01e97903b78
Create Date: 2017-05-23 16:16:28.941009

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '88942e83a048'
down_revision = 'b01e97903b78'


def upgrade():
    op.drop_column('energyprice', 'advance_pay')
    op.drop_column('energyprice', 'charge')


def downgrade():
    op.add_column(
        'energyprice',
        sa.Column('charge', sa.INTEGER(), autoincrement=False, nullable=True))
    op.add_column(
        'energyprice',
        sa.Column('advance_pay', sa.INTEGER(), autoincrement=False,
                  nullable=True))



# encoding=utf-8
from __future__ import unicode_literals
from sqlalchemy import Boolean
from sqlalchemy import Column, Integer, String, ForeignKey, Numeric, DateTime
from sqlalchemy import Date, Table
from sqlalchemy.orm import relationship, ColumnProperty
from sqlalchemy.orm.interfaces import AttributeExtension
from sqlalchemy.ext.instrumentation import InstrumentationManager
import datetime
import json
import pyramid.threadlocal
import risclog.sqlalchemy.interfaces
import risclog.sqlalchemy.model
import sqlalchemy.orm
import sys
import zope.component


ENGINE_NAME = 'portal'


class InstallValidatorListeners(InstrumentationManager):
    def post_configure_attribute(self, class_, key, inst):
        """Add validators for any attributes that can be validated."""
        prop = inst.prop
        # Only interested in simple columns, not relations
        if isinstance(prop, ColumnProperty) and len(prop.columns) == 1:
            col = prop.columns[0]
            if isinstance(col.type, String) and col.type.length:
                sqlalchemy.event.listen(
                    getattr(class_, key), 'set',
                    LengthValidator(col.type.length), retval=True)
            elif isinstance(col.type, Integer):
                sqlalchemy.event.listen(
                    getattr(class_, key), 'set',
                    IntegerValidator(), retval=True)


class ValidationError(Exception):
    pass


class LengthValidator(AttributeExtension):
    def __init__(self, max_length):
        self.max_length = max_length

    def __call__(self, state, value, oldvalue, initiator):
        if value is None:
            return value
        if value:
            value = str(value)
        if len(value) > self.max_length:
            raise ValidationError(
                "Länge von %d übersteigt erlaubte %d" % (
                    len(value), self.max_length))
        return value


class IntegerValidator(AttributeExtension):

    def __call__(self, state, value, oldvalue, initiator):
        if value is None or isinstance(value, int):
            return value
        try:
            return int(value)
        except Exception:
            raise ValidationError(
                "%s ist keine Zahl." % value)


class ObjectBase(risclog.sqlalchemy.model.ObjectBase):

    _engine_name = ENGINE_NAME
    __sa_instrumentation_manager__ = InstallValidatorListeners
    id = Column(Integer, primary_key=True)
    organization_id = Column(Integer, nullable=False)

    @classmethod
    def context_factory(cls, *args, **kw):
        result = super(ObjectBase, cls).context_factory(*args, **kw)
        # XXX importing browser here is a dependency in the wrong direction,
        # on the other hand, context_factory is used only by Pyramid
        import sw.allotmentclub.browser.auth
        result.__acl__ = sw.allotmentclub.browser.auth.authorize(
            context=result)
        return result

    @classmethod
    def create(cls, *args, **kw):
        if 'organization_id' not in kw:
            request = pyramid.threadlocal.get_current_request()
            if request and hasattr(request, 'user'):
                kw['organization_id'] = request.user.organization_id
            elif hasattr(sys, '_called_from_test'):
                kw['organization_id'] = 1
        return super(ObjectBase, cls).create(*args, **kw)

    def commit(self):
        db = zope.component.getUtility(risclog.sqlalchemy.interfaces.IDatabase)
        db.session.flush()


Object = risclog.sqlalchemy.model.declarative_base(
    ObjectBase, class_registry=risclog.sqlalchemy.model.class_registry)
MONEY = Numeric(12, 2)


class Organization(Object):
    """An organization like Leuna-Bungalowsiedlung."""

    organization_id = None
    title = Column(String)


class Log(Object):
    """Container in the database for security relevant incidents."""

    user_id = Column(Integer, ForeignKey('user.id'), nullable=False)
    name = Column(String(16))
    level = Column(String(8))
    created = Column(DateTime)
    msg = Column(String(2000))


ADDRESS = """
{appellation}<br />
{title} {firstname} {lastname}<br />
{street}<br />
{zip} {city}<br />
{country}"""

ADDRESS_CHANGE = {
    188: 187,  # Werkstatt -> Leuna Bungalowgemeinschaft
    82: 108,  # Britta Grimmling -> Andreas Grimmling
}


class PersonMixin(object):

    @property
    def address(self):
        address_data = json.loads(json.dumps(self))
        if self.id in ADDRESS_CHANGE:
            realto = Member.get(ADDRESS_CHANGE[self.id])
            realto_data = json.loads(json.dumps(realto))
            for data in ['street', 'zip', 'city', 'country']:
                address_data[data] = realto_data[data]
        address_data['country'] = (
            '' if address_data['country'] == 'Deutschland'
            else address_data['country'])
        return ADDRESS.format(**address_data)

    @property
    def deflection(self):
        return '' if self.appellation == 'Frau' else 'r'

    @property
    def greeting(self):
        return ('Sehr geehrte{deflection} {appellation} {title} '
                '{lastname},').format(**dict(
                    deflection=self.deflection,
                    appellation=self.appellation,
                    title=self.title,
                    lastname=self.lastname))


members_messages_table = Table(
    'members_messages', Object.metadata,
    Column('member_id', Integer, ForeignKey('member.id')),
    Column('message_id', Integer, ForeignKey('message.id'))
)


class Member(Object, PersonMixin):
    """A member in the club."""

    title = Column(String(20), default=u'')
    appellation = Column(String(20), default=u'')
    firstname = Column(String(50), default=u'')
    lastname = Column(String(50), default=u'')
    street = Column(String(100), default=u'')
    zip = Column(String(6), default=u'')
    city = Column(String(50), default=u'')
    country = Column(String(50), default=u'')
    phone = Column(String(50), default=u'')
    mobile = Column(String(50), default=u'')
    email = Column(String(100), default=u'')
    birthday = Column(Date)
    iban = Column(String(34), default=u'')
    bic = Column(String(11), default=u'')
    direct_debit_account_holder = Column(String(100))
    direct_debit = Column(Boolean)
    direct_debit_date = Column(DateTime)
    get_post = Column(Boolean, default=True)
    leaving_year = Column(Integer)

    messages = relationship("Message", secondary=members_messages_table)

    @property
    def assignment_hours(self):
        if self.id in (187, 188):  # Werkstatt, Verein
            return 5.0
        from .browser.base import get_selected_year
        hours = 0.0
        for assignmentattendee in self.assignments:
            if assignmentattendee.assignment.day.year == get_selected_year():
                hours += assignmentattendee.hours
        return hours

    def bill_assignment_hours(self):
        from .account import Booking, BookingKind
        missing_hours = 5 - self.assignment_hours
        from .browser.base import get_selected_year
        year = get_selected_year()
        kind = (BookingKind.query()
                .filter(BookingKind.shorttitle == 'NARB')).one()
        if missing_hours > 0:
            Booking.find_or_create(
                banking_account_id=2,
                booking_day=datetime.date(int(year), 10, 31),
                purpose='fehlende %s Arbeitsstunden %s' % (
                    missing_hours, year),
                value=0 - int(missing_hours * 10) * 10000,
                accounting_year=year,
                member=self,
                kind=kind)


class SaleHistory(Object):

    seller_id = Column(Integer, nullable=False)
    buyer_id = Column(Integer, nullable=False)
    date = Column(Date, nullable=False, default=datetime.date.today)


class Allotment(Object):
    """Die Parzelle."""

    number = Column(Integer, nullable=False)
    member_id = Column(Integer, ForeignKey(Member.id), nullable=False)
    member = sqlalchemy.orm.relation(
        'Member', uselist=False, backref='allotments')


class Parcel(Object):
    """Das Flustück."""

    number = Column(Integer, nullable=False)
    leased = Column(Boolean, nullable=False, default=False)
    tap_water = Column(Boolean, nullable=False, default=False)
    allotment_id = Column(Integer, ForeignKey(Allotment.id), nullable=False)
    allotment = sqlalchemy.orm.relation(
        'Allotment', uselist=False, backref='parcels')


class AccessAuthority(Object):
    """Eine Zugriffssteuerung."""

    viewname = Column(String)
    user_id = Column(Integer, ForeignKey('user.id'), nullable=True)
    user = sqlalchemy.orm.relation('User', uselist=False)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import pkg_resources
import datetime
import transaction


def setUp():
    from sw.allotmentclub import (
        Protocol, ProtocolDetail, ProtocolCommitment, ProtocolAttachment)
    protocol = Protocol.create(
        day=datetime.datetime(2015, 3, 7, 10),
        title='1. Vorstandssitzung',
        location='Leuna Siedling, Vereinsbungalow',
        attendees='GM, AR, ST, SW',
        accounting_year=2015)
    ProtocolDetail.create(
        protocol=protocol,
        duration='10',
        message="""
**TOP I: Berichte**

- GM berichtet über dies.
- ST über jenes.
- dabei fällt SW auf, dass AR garnicht anwesend ist.

Na sowas.""",
        responsible='GM')
    ProtocolDetail.create(
        protocol=protocol,
        duration='20',
        message="""
**TOP II: Finanzen**

Geld ist genug da. (Sagt ST)""",
        responsible='ST')
    ProtocolCommitment.create(
        protocol=protocol,
        who='SW',
        what='Verwaltungs-Software endlich fertig bekommen',
        when='Mitte 2015')
    ProtocolCommitment.create(
        protocol=protocol,
        who='ST',
        what='Mehr Geld ausgeben',
        when='Ende 2015')
    img = get_test_image()
    ProtocolAttachment.create(
        protocol=protocol,
        name='Anlage I',
        mimetype='image/gif',
        size=len(img),
        data=img)
    transaction.commit()


def get_test_image():
    return pkg_resources.resource_stream(
        'sw.allotmentclub.browser.tests', 'assyrian.gif').read()


def test_protocol_can_have_details(database):
    from sw.allotmentclub import Protocol
    setUp()
    protocol = Protocol.query().one()
    assert 2 == len(protocol.details)
    assert 2 == len(protocol.commitments)
    assert 1 == len(protocol.attachments)


def test_displays_list_of_protocols(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/protocols?for_year=2015')
    expected = [1, '07.03.2015 10:00', '1. Vorstandssitzung',
                'GM, AR, ST, SW', 'Leuna Siedling, Vereinsbungalow']
    assert expected in browser.json_result


def test_protocol_can_be_added_via_json_view(browser):
    from sw.allotmentclub import Protocol
    setUp()
    browser.login()
    browser.post('http://localhost/protocols/add', data='')
    assert 'success' == browser.json['status']
    assert 2 == len(Protocol.query().all())


def test_displays_list_of_protocol_details(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/protocols/1/details')
    assert [2, 20, '<p><strong>TOP II: Finanzen</strong></p>\n<p>Geld ist '
            'genug da. (Sagt ST)</p>', 'ST'] in browser.json_result


def test_protocols_can_be_printed_as_pdf(browser):
    setUp()
    browser.login()
    browser.open('http://localhost/protocols/1/print')
    assert browser.contents.decode('latin1').startswith('%PDF-')


# class IProtocolCRUDTest(sw.allotmentclub.browser.testing.SeleniumTestCase):
#
#     @mock.patch('sw.allotmentclub.browser.protocol.datetime_now')
#     def test_protocol_crud(self, now):
#         current_year = datetime.datetime.now().year
#         now.return_value = datetime.datetime(current_year, 2, 20)
#         sel = self.selenium
#         sel.waitForTextPresent('Protokolle')
#         sel.click('link=Protokolle')
#         self.assertEmptyTable()
#         # Protokoll anlegen
#         sel.click('link=Neu')
#         sel.waitForElementPresent('css=input[name=day]')
#         self.enter('input[name=title]', 'Protokoll Vorstandssitzung')
#         self.enter('input[name=attendees]', 'GM, AR, ST, SW')
#         self.enter('input[name=location]', 'Roter See, Vereinsbungalow')
#         sel.click('css=#sub_content button[type=submit]')
#         self.markLineInTable()
#
#         # Protokoll bearbeiten
#         sel.click('link=Bearbeiten')
#         sel.waitForElementPresent('css=input[name=day]')
#         sel.assertValue(
#             'css=input[name=title]', 'Protokoll Vorstandssitzung')
#         self.enter('input[name=attendees]', ', HQ')
#         sel.click('css=#sub_content button[type=submit]')
#         sel.waitForTextPresent('GM, AR, ST, SW, HQ')
#         self.markLineInTable()
#
#         # Protokolldetail hinzufügen
#         sel.click('link=Details')
#         self.assertEmptyTable()
#         sel.click('css=#sub_content a[href=protocol_detail_add]')
#         sel.waitForElementPresent('css=input[name=duration]')
#         self.enter('input[name=duration]', '5')
#         self.enter_markdown(
#             'textarea.md-input', '**TOP I**\n\nBegrüßung der Teilnehmer')
#         self.enter('input[name=responsible]', 'GM')
#         sel.click('css=#sub_content button[type=submit]')
#         self.assertNotEmptyTable()
#         sel.assertText('css=tr.odd td p strong', 'TOP I')
#         self.markLineInTable()
#
#         # Protokolldetail bearbeiten
#         sel.click('link=Details')
#         self.markLineInTable(subcontent=True)
#         sel.click('css=#sub_content a[href=protocol_detail_edit]')
#         sel.waitForElementPresent('css=input[name=duration]')
#         self.enter('input[name=duration]', '0')
#         sel.click('css=#sub_content button[type=submit]')
#         sel.waitForTextPresent('50')
#         self.markLineInTable()
#
#         # Anlagen hinzufügen
#         sel.click('link=Anlagen')
#         self.assertEmptyTable()
#         sel.click('css=#sub_content .btn-success')
#         sel.waitForElementPresent('css=.dz-clickable')
#         self.upload_dz(pkg_resources.resource_filename(
#             'sw.allotmentclub.browser.tests', 'example.gif'))
#         self.assertNotEmptyTable()
#         sel.waitForTextPresent('Anlage I')
#         sel.waitForTextPresent('GIF')
#         sel.waitForTextPresent('1.35 KB')
#
#         # Commitments hinzufügen
#         sel.click('link=Absprachen')
#         self.assertEmptyTable()
#         sel.click('css=#sub_content a[href=protocol_commitment_add]')
#         sel.waitForElementPresent('css=input[name=who]')
#         self.enter('input[name=who]', 'SW')
#         self.enter('input[name=what]', 'Nächste Besprechung vorbereiten')
#         self.enter('input[name=when]', 'KW14')
#         sel.click('css=#sub_content button[type=submit]')
#         self.assertNotEmptyTable()
#
#         # Commitments bearbeiten
#         sel.click('link=Absprachen')
#         self.markLineInTable(subcontent=True)
#         sel.click('css=#sub_content a[href=protocol_commitment_edit]')
#         sel.waitForElementPresent('css=input[name=who]')
#         self.enter('input[name=who]', 'W')
#         sel.click('css=#sub_content button[type=submit]')
#         sel.waitForTextPresent('SWW')
#
#         # PDF laden
#         self.assertPDFEqual(
#             'api/protocols/1/print', 'test_protocol_print.pdf')
#
#         # Protokolldetail löschen
#         sel.click('link=Details')
#         self.markLineInTable(subcontent=True)
#         sel.click('css=#sub_content .btn-danger')
#         self.confirm_delete()
#         self.assertEmptyTable()
#
#         # Commitments löschen
#         sel.click('link=Absprachen')
#         self.markLineInTable(subcontent=True)
#         sel.click('css=#sub_content .btn-danger')
#         self.confirm_delete()
#         self.assertEmptyTable()



"""Update messages to integrate greeting into text.

Revision ID: 2330cb148c04
Revises: ee1b8c9cb745
Create Date: 2016-04-25 10:34:10.696062

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '2330cb148c04'
down_revision = 'ee1b8c9cb745'


def upgrade():
    op.execute(
        """UPDATE message
           SET body = 'Sehr geehrte{{deflection}} {{appellation}}
           {{title}} {{lastname}},\n\n' || body;""")


def downgrade():
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from ..log import user_data_log, log_with_user
from .base import date_time, format_size, format_mimetype
from io import BytesIO
from pyramid.response import FileIter
from pyramid.view import view_config
from sw.allotmentclub import Depot, User
import datetime
import sw.allotmentclub.browser.base


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Größe': format_size,
        'Dateityp': format_mimetype,
        'Aktualisiert': date_time,
    }

    data_class = {
        'Dateiname': 'expand'
    }
    data_hide = {
        'Dateityp': 'phone,tablet',
        'Größe': 'phone,tablet',
        'Aktualisiert': 'phone,tablet',
        'Von': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                Depot.id.label('#'),
                Depot.name.label('Dateiname'),
                Depot.mimetype.label('Dateityp'),
                Depot.size.label('Größe'),
                Depot.date.label('Aktualisiert'),
                User.username.label('Von'))
            .select_from(Depot)
            .join(User))


@view_config(route_name='depots', renderer='json',
             permission='view')
class DepotListView(sw.allotmentclub.browser.base.TableView):
    """ Liste aller Depots."""

    query_class = Query
    default_order_by = 'name'
    available_actions = [
        dict(url='depot_add', btn_class='btn-success', icon='fa fa-plus',
             title='Neu'),
        dict(url='depot_edit', btn_class='btn-success', icon='fa fa-pencil',
             title='Bearbeiten'),
        dict(url='depot_delete', btn_class='btn-danger',
             icon='glyphicon glyphicon-trash', title='Löschen'),
        dict(url='depot_download', btn_class='btn-success',
             icon='fa fa-download', title='Herunterladen')]


class DepotBase(sw.allotmentclub.browser.base.View):

    def parse_file(self):
        file = self.request.params.get('file')
        data = file.file
        data.seek(0)
        data = data.read()
        name = file.filename
        mimetype = file.type
        size = len(data)
        date = datetime.datetime.now()
        user = self.request.user
        return name, mimetype, size, data, date, user

    def log(self, id):
        log_with_user(user_data_log.info,
                      self.request.user,
                      'hat Ablage %s %s.', id, self.action)


@view_config(route_name='depot_add', renderer='json',
             permission='view')
class DepotAddView(DepotBase):

    action = 'hinzugefügt'

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data, date, user = self.parse_file()
        depot = Depot.create(
            name=name, mimetype=mimetype, size=size, data=data,
            date=date, user=user)
        depot.commit()
        self.log(depot.id)
        return {'status': 'success'}


@view_config(route_name='depot_edit', renderer='json',
             permission='view')
class DepotEditView(DepotBase):

    action = 'bearbeitet'

    def __call__(self):
        if not self.form_submit():
            return {'status': 'success', 'data': {}}
        name, mimetype, size, data, date, user = self.parse_file()
        if mimetype != self.context.mimetype:
            return {'status': 'error',
                    'message': 'Kann nur Dateien vom gleichen Typ bearbeiten. '
                               'alt: {}, neu: {}'.format(
                                   self.context.mimetype, mimetype)}
        self.context.size = size
        self.context.data = data
        self.context.date = date
        self.context.name = name
        self.context.user = user
        self.log(self.context.id)
        return {'status': 'success'}


@view_config(route_name='depot_delete', renderer='json',
             permission='view')
class DepotDeleteView(sw.allotmentclub.browser.base.DeleteView):

    model = Depot


@view_config(route_name='depot_download',
             permission='view')
class DepotDownloadView(object):

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = self.context.mimetype
        response.content_length = int(self.context.size)
        response.content_disposition = (
            'attachment; filename=%s' % self.context.name)
        response.app_iter = FileIter(BytesIO(self.context.data))
        return response



# encoding=utf8
from __future__ import unicode_literals
from ..log import user_data_log, log_with_user
from .base import date, get_selected_year, value_to_int, boolean, date_time
from .base import format_eur_with_color, format_eur, to_string, string_agg
from .mail import append_pdf, render_pdf, format_markdown
from PyPDF2 import PdfFileWriter, PdfFileReader
from io import BytesIO
from pyramid.response import FileIter
from pyramid.view import view_config
from sqlalchemy.sql import func
from sw.allotmentclub import Booking, Member, Allotment, GrundsteuerB, Abwasser
from sw.allotmentclub import BookingKind, SEPASammler, SEPASammlerEntry
from sw.allotmentclub import VALUE_PER_MEMBER
import collections
import datetime
import dateutil.parser
import pybars
import re
import sw.allotmentclub.browser.base

BOOKING_FUTURE = datetime.datetime.now() + datetime.timedelta(days=61)


def AccountDetailFactory(request):
    account_id, account_type = request.matchdict['id'].split('-')
    context = getattr(sw.allotmentclub.account,
                      account_type).get(account_id)
    context.__acl__ = sw.allotmentclub.browser.auth.DEFAULT_ACL
    return context


class Query(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Betrag': format_eur_with_color,
    }

    css_classes = {
        'Betrag': 'right nowrap',
        'Datum': 'nowrap',
        'Zugeordnet': 'nowrap',
    }

    data_class = {
        'Buchungsinfo': 'expand'
    }
    data_hide = {
        'Datum': 'phone',
        'Betrag': 'phone',
        'Zugeordnet': 'phone,tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import false
        return (
            self.db.query(
                Booking.id.label('#'),
                Booking.booking_day.label('Datum'),
                (to_string(Booking.booking_text).concat('<br />')
                 .concat(to_string(Booking.purpose)).concat('<br />')
                 .concat(to_string(Booking.recipient)).label('Buchungsinfo')),
                Booking.value.label('Betrag'),
                (to_string(Member.lastname).concat(' (')
                 .concat(func.coalesce(string_agg(Allotment.number), 'n/a'))
                 .concat(')').label('Zugeordnet')),
                BookingKind.shorttitle.label('Typ'),
            )
            .select_from(Booking)
            .outerjoin(Member)
            .outerjoin(Allotment)
            .outerjoin(BookingKind)
            .group_by(Booking.id, Member.lastname, BookingKind.shorttitle)
            .filter(Booking.banking_account == self.context)
            .filter(Booking.is_splitted == false())
            .filter(Booking.accounting_year == get_selected_year()))


@view_config(route_name='booking_list', renderer='json',
             permission='view')
class BookingListView(sw.allotmentclub.browser.base.TableView):

    query_class = Query
    default_order_by = 'booking_day'
    year_selection = True
    available_actions = [
        dict(url='map_booking', btn_class='btn-success',
             icon='fa fa-hand-o-right', title='Zuordnen'),
        dict(url='split_booking', btn_class='btn-success',
             icon='fa fa-scissors', title='Aufteilen')
    ]


@view_config(route_name='split_booking', renderer='json',
             permission='view')
class SplitBookingView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Buchung aufteilen'

    @property
    def load_options(self):
        return {
            'new_value': {
                'label': 'Teilen bei Betrag',
                'required': True
            }
        }

    @property
    def load_data(self):
        fields = [
            ('new_value', '')
        ]
        return collections.OrderedDict(fields)

    def fail(self, message):
        self.result['status'] = 'error'
        self.result['msg'] = message

    def update(self):
        try:
            json = self.request.json
        except ValueError:
            json = None
        if json:
            new_value = value_to_int(json['new_value'])
            status = self.context.split(new_value)
            if isinstance(status, str):
                self.fail(status)
                return
            log_with_user(
                user_data_log.info, self.request.user,
                'Buchung %s bei %s aufgeteilt.',
                self.context.id, format_eur(new_value))
            self.result = {'status': 'success'}
        else:
            self.result = self.get_result()


@view_config(route_name='map_booking', renderer='json',
             permission='view')
class MapBookingView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'Buchung einem Mitglied zuweisen'

    @property
    def load_options(self):
        result = {
            'member_id': {
                'source': self.member_source,
                'label': 'Mitglied',
                'css_class': 'chosen',
                'required': False
            },
            'kind_id': {
                'source': self.booking_kind_source,
                'label': 'Typ',
                'css_class': 'chosen',
                'required': False
            },
            'accounting_year': {
                'source': self.accounting_year_source,
                'label': 'Jahr',
                'css_class': 'chosen',
                'required': True
            }
        }
        return result

    @property
    def load_data(self):
        fields = [
            ('member_id', self.context.member_id),
            ('kind_id', self.context.kind_id),
            ('accounting_year', self.context.accounting_year)]
        return collections.OrderedDict(fields)


def get_balance(value, request=None):
    from sqlalchemy.sql.expression import false
    import zope.component
    import risclog.sqlalchemy
    member = Member.get(value)
    db = zope.component.getUtility(risclog.sqlalchemy.interfaces.IDatabase)
    bookings = (
        db.query(func.sum(Booking.value))
        .join(Member)
        .filter(Booking.member == member)
        .filter(Booking.is_splitted == false())
        .filter(Booking.booking_day <= BOOKING_FUTURE)
        .filter(Booking.accounting_year == get_selected_year()))
    sepa = (
        db.query(func.sum(SEPASammlerEntry.value))
        .join(Member)
        .join(SEPASammler)
        .filter(SEPASammlerEntry.member == member)
        .filter(SEPASammler.booking_day <= BOOKING_FUTURE)
        .filter(SEPASammler.accounting_year == get_selected_year()))
    booking_sum = bookings.one()[0]
    sepa_sum = sepa.one()[0]
    booking_sum = 0 if booking_sum is None else booking_sum
    sepa_sum = 0 if sepa_sum is None else sepa_sum
    return format_eur_with_color(booking_sum + sepa_sum)


class MemberAccountQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Konto': get_balance,
    }

    css_classes = {
        'Konto': 'right nowrap',
    }

    data_class = {
        'Mitglied': 'expand'
    }
    data_hide = {
        'Konto': 'phone',
    }

    def select(self):
        return (
            self.db.query(
                Member.id.label('#'),
                (to_string(Member.lastname).concat(', ')
                 .concat(to_string(Member.firstname)).label('Mitglied')),
                Member.id.label('Konto')
            )
            .select_from(Member))


@view_config(route_name='member_account_list', renderer='json',
             permission='view')
class MemberAccountListView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberAccountQuery
    default_order_by = 'member.id'
    year_selection = True
    available_actions = [
        dict(url='member_account_detail_list', btn_class='btn-success',
             icon='fa fa-list', title='Buchungen'),
    ]


class MemberAccountDetailQuery(sw.allotmentclub.browser.base.Query):

    disable_global_organization_filter = True

    formatters = {
        'Betrag': format_eur_with_color,
        'Datum': date,
        'RS': boolean
    }

    css_classes = {
        'Betrag': 'right nowrap',
    }

    data_class = {
        'Datum': 'expand'
    }
    data_hide = {
        'Betrag': 'phone',
        'Typ': 'phone, tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import false
        bookings = (
            self.db.query(
                to_string(Booking.id).concat('-Booking').label('#'),
                Booking.booking_day.label('Datum'),
                BookingKind.title.label('Typ'),
                Booking.value.label('Betrag'),
                Booking.ignore_in_reporting.label('RS')
            )
            .select_from(Booking)
            .outerjoin(BookingKind)
            .join(Member)
            .filter(Booking.is_splitted == false())
            .filter(Booking.accounting_year == get_selected_year())
            .filter(Booking.booking_day <= BOOKING_FUTURE)
            .filter(Booking.organization_id == self.user.organization_id)
            .filter(Booking.member == self.context))
        sepa = (
            self.db.query(
                (to_string(SEPASammlerEntry.id)
                 .concat('-SEPASammlerEntry').label('#')),
                SEPASammler.booking_day.label('Datum'),
                BookingKind.title.label('Typ'),
                SEPASammlerEntry.value.label('Betrag'),
                SEPASammlerEntry.ignore_in_reporting.label('RS')
            )
            .select_from(SEPASammler)
            .join(SEPASammlerEntry)
            .outerjoin(BookingKind)
            .join(Member)
            .filter(SEPASammler.accounting_year == get_selected_year())
            .filter(SEPASammler.booking_day <= BOOKING_FUTURE)
            .filter(SEPASammler.organization_id == self.user.organization_id)
            .filter(SEPASammlerEntry.member == self.context))
        return bookings.union(sepa)


@view_config(route_name='member_account_detail_list', renderer='json',
             permission='view')
class MemberAccountDetailListView(sw.allotmentclub.browser.base.TableView):

    query_class = MemberAccountDetailQuery
    year_selection = True
    available_actions = [
        dict(url='member_account_detail_switch_ir',
             btn_class='btn-danger',
             icon='fa fa-check-square-o ', title='Reporting-Status ändern'),
    ]


@view_config(route_name='member_account_detail_switch_ir', renderer='json',
             permission='view')
class MemberAccountDetailSwitchRSView(sw.allotmentclub.browser.base.View):

    def update(self):
        self.context.ignore_in_reporting = not self.context.ignore_in_reporting
        self.result = {'status': 'success',
                       'message': 'Reporting-Status geändert.'}


def get_outstanding_raw(value, request=None, year=None):
    from sqlalchemy.sql.expression import false
    year = year if year else get_selected_year()
    kind = BookingKind.get(value)
    if kind.id == 19:
        return (
            get_outstanding_raw(3) +
            get_outstanding_raw(4) +
            get_outstanding_raw(5))
    else:
        query = (
            Booking.query()
            .join(Member)
            .filter(Booking.kind == kind)
            .filter(Booking.accounting_year == year)
            .filter(Booking.ignore_in_reporting == false())
            .filter(Booking.is_splitted == false())
            .filter(Booking.banking_account_id != 1)
        )
        return query.all()


def get_outstanding_sum(value, request=None, year=None):
    kind = BookingKind.get(value)
    sum_ = sum(b.value for b in get_outstanding_raw(value, request, year))
    if kind.id == 19:
        return 0 - sum_
    return sum_


def get_outstanding(value, request=None, year=None):
    return format_eur_with_color(get_outstanding_sum(value, request, year))


def get_incoming_raw(value, request=None, year=None):
    from sqlalchemy.sql.expression import false
    year = year if year else get_selected_year()
    kind = BookingKind.get(value)
    query = (
        Booking.query()
        .join(Member)
        .filter(Booking.kind == kind)
        .filter(Booking.accounting_year == year)
        .filter(Booking.ignore_in_reporting == false())
        .filter(Booking.is_splitted == false())
        .filter(Booking.banking_account_id == 1)
        .filter(Booking.booking_text != 'SAMMEL-LS-EINZUG')
    )
    sammler = (
        SEPASammlerEntry.query()
        .join(SEPASammler)
        .join(Member)
        .filter(SEPASammler.kind == kind)
        .filter(SEPASammlerEntry.ignore_in_reporting == false())
        .filter(SEPASammler.accounting_year == year)
    )
    return query.all() + sammler.all()


def get_incoming_sum(value, request=None, year=None):
    return sum(b.value for b in get_incoming_raw(value, request, year))


def get_incoming(value, request=None, year=None):
    return format_eur_with_color(get_incoming_sum(value, request, year))


def get_incoming_pre_year(value, request=None):
    return get_incoming(value, request, get_selected_year()-1)


def get_sum(value, request=None):
    return format_eur_with_color(
        get_outstanding_sum(value, request) +
        get_incoming_sum(value, request))


class BankingAccountQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Vorjahr': get_incoming_pre_year,
        'Soll': get_outstanding,
        'Ist': get_incoming,
        'Offen': get_sum,
    }

    css_classes = {
        'Vorjahr': 'right nowrap',
        'Soll': 'right nowrap',
        'Ist': 'right nowrap',
        'Offen': 'right nowrap',
    }

    data_class = {
        'Kategorie': 'expand'
    }
    data_hide = {
        'Vorjahr': 'phone',
        'Soll': 'phone',
        'Ist': 'phone',
        'Offen': 'phone,tablet'
    }

    def select(self):
        return (
            self.db.query(
                BookingKind.id.label('#'),
                BookingKind.title.label('Kategorie'),
                BookingKind.id.label('Vorjahr'),
                BookingKind.id.label('Soll'),
                BookingKind.id.label('Ist'),
                BookingKind.id.label('Offen')
            )
            .select_from(BookingKind))


@view_config(route_name='banking_account_list', renderer='json',
             permission='view')
class BankingAccountListView(sw.allotmentclub.browser.base.TableView):

    query_class = BankingAccountQuery
    year_selection = True
    available_actions = [
        dict(url='banking_account_list_detail', btn_class='btn-success',
             icon='fa fa-list', title='Positionen'),
        dict(url='banking_account_list_report', btn_class='btn-success',
             icon='fa fa-print', title='Report')
    ]


@view_config(route_name='banking_account_list_detail', renderer='json',
             permission='view')
class BankingAccountListDetailView(sw.allotmentclub.browser.base.TableView):

    def __call__(self):
        incoming = get_incoming_raw(self.context.id, self.request)
        outstanding = get_outstanding_raw(self.context.id, self.request)
        columns = [
            {'name': name, 'css_class': 'hide' if id == 0 else ''}
            for id, name in enumerate(
                ['#', 'Betrag', 'Mitglied', 'Buchungstext', 'TYP'])]
        result = []
        for type_, objs in enumerate([incoming, outstanding]):
            for obj in objs:
                value = obj.value
                if self.context.id == 19 and type_ == 1:
                    value = 0 - value
                new_line = [
                    {'value': obj.id, 'css_class': 'hide'},
                    {'value': format_eur_with_color(value, self.request)},
                    {'value': '{}, {} ({})'.format(
                        obj.member.lastname, obj.member.firstname,
                        (obj.member.allotments[0].number
                         if obj.member.allotments else ''))},
                    {'value': (obj.purpose if hasattr(obj, 'booking_text')
                               else 'Sammler-LS')},
                    {'value': 'SOLL' if type_ == 1 else 'IST'},
                ]
                result.append(new_line)
        data = {'data': result,
                'header': columns,
                'actions': [],
                'records': len(result)}
        return {'status': 'success', 'data': data}


@view_config(route_name='banking_account_list_report', renderer='json',
             permission='view')
class BankingAccountListReportView(sw.allotmentclub.browser.base.View):

    hbs_page = """
<h3>{{title}}</h3>
<table style="border: 1px solid black;">
  <tr style="font-style: normal; font-size: 8pt;">
    <th style="width: 15%; {{{header_style}}}">Betrag</th>
    <th style="width: 25%; {{{header_style}}}">Mitglied</th>
    <th style="width: 50%; {{{header_style}}}">Buchungstext</th>
    <th style="width: 10%; {{{header_style}}}">Typ</th>
  </tr>
  {{#each content}}
    <tr>
      {{#each this}}
      <td style="border: 1px solid black;
                 padding: 2px 2px -1px 2px;
                 {{#unless @index}} text-align: right; {{/unless}}
                 font-style: normal;
                 font-size: 8pt;">{{this}}</td>
      {{/each}}
    </tr>
  {{/each}}
  <tr style="font-style: normal; font-size: 8pt;">
    <th style="width: 15%; {{{header_style}}}">SUMME</th>
    <th colspan="3" style="{{{header_style}}}">{{summe}}</th>
  </tr>
</table>
"""

    def update(self):
        output = PdfFileWriter()
        compiler = pybars.Compiler()
        compiled_page = compiler.compile(self.hbs_page)
        cleanr = re.compile('<.*?>')

        for kind in BookingKind.query():
            data = BankingAccountListDetailView(kind, self.request)()['data']
            table_content = []
            for item in data['data']:
                table_content.append(
                    [re.sub(cleanr, '', d['value']) for d in item[1:]])
            table_content = sorted(
                table_content, key=lambda x: x[1])  # sort by name
            sum_ = sum(float(
                i[0].replace(' €', '').replace('.', '').replace(',', '.'))
                for i in table_content)
            sum_ = int(sum_ * 10000)

            body = format_markdown(compiled_page(dict(
                title='{} {}'.format(
                    kind.title, get_selected_year()),
                content=table_content,
                summe=format_eur(sum_),
                header_style=("border: 1px solid black; "
                              "padding: 2px 2px -1px 2px;"))))
            output = self.add_page(body, output)
        result = BytesIO()
        output.write(result)
        result.seek(0)
        response = self.request.response
        response.set_cookie('fileDownload', value='true')
        response.content_type = 'application/pdf'
        response.content_disposition = 'attachment; filename=Report 2018.pdf'
        response.app_iter = FileIter(result)
        self.result = response

    def add_page(self, body, output):
        pdf = render_pdf(None, None, body, None, date=None, force_from=False)
        pdf = PdfFileReader(pdf, strict=False)
        append_pdf(pdf, output)
        return output


class SEPASammlerQuery(sw.allotmentclub.browser.base.Query):

    data_class = {
        'Art': 'expand'
    }
    data_hide = {
        'Sparkassen-ID': 'phone,tablet',
        'Datum': 'phone,tablet',
    }

    def select(self):
        return (
            self.db.query(
                SEPASammler.id.label('#'),
                BookingKind.title.label('Art'),
                SEPASammler.booking_day.label('Datum'),
                SEPASammler.pmtinfid.label('Sparkassen-ID'),
            )
            .select_from(SEPASammler)
            .outerjoin(BookingKind)
            .filter(SEPASammler.accounting_year == get_selected_year()))


@view_config(route_name='sepa_sammler_list', renderer='json',
             permission='view')
class SEPASammlerListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPASammlerQuery
    year_selection = True
    start_year = 2015
    available_actions = [
        dict(url='sepa_sammler_add', btn_class='btn-success',
             icon='fa fa-plus', title='Neu'),
        dict(url='sepa_sammler_edit', btn_class='btn-success',
             icon='fa fa-pencil', title='Bearbeiten'),
        dict(url='sepa_sammler_entry_list', btn_class='btn-success',
             icon='fa fa-list', title='Positionen'),
        dict(url='sepa_sammler_update', btn_class='btn-success',
             icon='fa fa-list', title='autom. Befüllen'),
        dict(url='sepa_sammler_export', btn_class='btn-success',
             icon='fa fa-bank', title='Export Sparkasse'),
    ]


@view_config(route_name='sepa_sammler_update', renderer='json',
             permission='view')
class SEPASammlerUpdateView(sw.allotmentclub.browser.base.View):

    def advancepay(self):
        values = (
            Booking.query()
            .filter(Booking.accounting_year == self.context.accounting_year)
            .filter(Booking.kind == self.context.kind).all())
        for value in values:
            if value.value < 0:
                if value.member and value.member.direct_debit:
                    SEPASammlerEntry().find_or_create(
                        sepasammler=self.context,
                        value=0 - value.value,
                        member=value.member)

    def energieabrechnung(self):
        values = (
            Booking.query()
            .filter(Booking.accounting_year == self.context.accounting_year)
            .filter(Booking.kind == self.context.kind)
            .filter(Booking.value <= 0).all())
        for value in values:
            if value.member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=0 - value.value,
                    member=value.member)

    def grundsteuerb(self):
        values = GrundsteuerB.query().all()
        for value in values:
            member = value.parcel.allotment.member
            if member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value.value,
                    member=value.parcel.allotment.member)

    def abwasser(self):
        values = Abwasser.query().all()
        for value in values:
            member = value.parcel.allotment.member
            if member.direct_debit:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value.value,
                    member=member)

    def arbeitsstunden(self):
        members = (Member.query()
                         .join(Allotment)
                         .filter(Member.direct_debit.is_(True)))
        for member in members.all():
            bookings = (
                Booking.query()
                .filter(
                    Booking.accounting_year == self.context.accounting_year)
                .filter(Booking.kind == self.context.kind)
                .filter(Booking.member == member).all())
            value = 0
            for booking in bookings:
                value -= booking.value
            if value:
                SEPASammlerEntry().find_or_create(
                    sepasammler=self.context,
                    value=value,
                    member=member)

    def mitgliedsbeitrag(self):
        members = Member.query().join(Allotment).filter(
            Member.direct_debit.is_(True))
        for member in members.all():
            SEPASammlerEntry().find_or_create(
                sepasammler=self.context,
                value=VALUE_PER_MEMBER,
                member=member)

    def update(self):
        if not self.context.kind:
            self.result = {'status': 'error',
                           'message': 'Art muss gesetzt sein.'}
            return
        for entry in (
                SEPASammlerEntry.query()
                .filter(SEPASammlerEntry.sepasammler == self.context).all()):
            entry.delete()
        if self.context.kind.shorttitle in ('ENA1', 'ENA2'):
            self.advancepay()
        elif self.context.kind.shorttitle == 'ENAB':
            self.energieabrechnung()
        elif self.context.kind.shorttitle == 'GRUB':
            self.grundsteuerb()
        elif self.context.kind.shorttitle == 'ABWA':
            self.abwasser()
        elif self.context.kind.shorttitle == 'NARB':
            self.arbeitsstunden()
        elif self.context.kind.shorttitle == 'BEIT':
            self.mitgliedsbeitrag()
        else:
            self.result = {
                'status': 'error',
                'message': '{} nicht unterstützt'.format(
                    self.context.kind.title)
            }
            return
        self.result = {
            'status': 'success',
            'message': 'SEPASammler {} berechnet.'.format(
                self.context.kind.title)}


@view_config(route_name='sepa_sammler_edit', renderer='json',
             permission='view')
class SEPASammlerEditView(sw.allotmentclub.browser.base.EditJSFormView):

    title = 'SEPA Sammler bearbeiten'

    @property
    def load_options(self):
        return {
            'kind_id': {
                'label': 'Art',
                'css_class': 'chosen',
                'required': True,
                'source': self.booking_kind_source,
            },
            'booking_day': {
                'label': 'Datum',
                'css_class': 'datetimepicker',
            },
            'pmtinfid': {
                'label': 'Sparkassen-ID',
            }
        }

    @property
    def load_data(self):
        fields = [
            ('kind_id', self.context.kind_id),
            ('booking_day', date_time(self.context.booking_day)),
            ('pmtinfid', self.context.pmtinfid),
        ]
        return collections.OrderedDict(fields)

    def save(self, key, value):
        if key == 'booking_day' and value:
            value = dateutil.parser.parse(value).date()
        return super(SEPASammlerEditView, self).save(key, value)

    def update(self):
        super(SEPASammlerEditView, self).update()
        if self.context.booking_day:
            if isinstance(self.context.booking_day, datetime.date):
                self.context.accounting_year = self.context.booking_day.year
            else:
                self.context.accounting_year = (
                    dateutil.parser.parse(self.context.booking_day).year)


@view_config(route_name='sepa_sammler_add', renderer='json',
             permission='view')
class SEPASammlerAddView(SEPASammlerEditView):

    def __init__(self, context, request):
        context = SEPASammler.create()
        context.commit()
        super(SEPASammlerAddView, self).__init__(context, request)
        log_with_user(
            user_data_log.info, self.request.user,
            'SEPA Sammler %s hinzugefügt.', self.context.id)

    @property
    def route_name(self):
        return 'sepa_sammler_edit'


@view_config(route_name='sepa_sammler_export', renderer='json',
             permission='view')
class SEPASammlerExportView(sw.allotmentclub.browser.base.XMLExporterView):

    @property
    def filename(self):
        return '{} {}'.format(
            self.context.kind.title, self.context.accounting_year)

    @property
    def subject(self):
        return self.filename

    @property
    def values(self):
        return sorted(self.context.entries, key=lambda x: x.member.lastname)

    def get_member(self, value):
        return value.member

    def get_to_pay(self, value):
        return value.value

    @property
    def creation_date(self):
        return datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')

    @property
    def booking_day(self):
        return self.context.booking_day.strftime('%Y-%m-%d')


class SEPASammlerEntryQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Betrag': format_eur,
        'Ignorieren': boolean,
    }

    data_class = {
        'Zugeordnet': 'expand'
    }

    data_hide = {
        'Ignorieren': 'phone,tablet',
        'Betrag': 'phone'
    }

    def select(self):
        return (
            self.db.query(
                SEPASammlerEntry.id.label('#'),
                (to_string(Member.lastname).concat(', ').concat(
                 to_string(Member.firstname).concat(' (')
                 .concat(func.coalesce(string_agg(Allotment.number), 'n/a'))
                 .concat(')'))).label('Zugeordnet'),
                SEPASammlerEntry.value.label('Betrag'),
                SEPASammlerEntry.ignore_in_reporting.label('Ignorieren')
            )
            .select_from(SEPASammlerEntry)
            .filter(SEPASammlerEntry.sepasammler == self.context)
            .join(Member)
            .outerjoin(Allotment)
            .group_by(SEPASammlerEntry.id, Member.lastname, Member.firstname))


@view_config(route_name='sepa_sammler_entry_list', renderer='json',
             permission='view')
class SEPASammlerEntryListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPASammlerEntryQuery
    year_selection = False


def account_holder(id, request=None):
    member = Member.get(id)
    result = member.direct_debit_account_holder or (
        '%s, %s' % (member.lastname, member.firstname))
    return '%s (%s)' % (
        result, '/'.join(str(a.number) for a in member.allotments))


class SEPADirectDebitQuery(sw.allotmentclub.browser.base.Query):

    formatters = {
        'Mitglied': account_holder,
    }
    data_class = {
        'Mitglied': 'expand'
    }
    data_hide = {
        'IBAN': 'phone',
        'BIC': 'phone,tablet',
        'Datum': 'phone,tablet',
    }

    def select(self):
        from sqlalchemy.sql.expression import true
        return (
            self.db.query(
                Member.id.label('#'),
                Member.id.label('Mitglied'),
                Member.iban.label('IBAN'),
                Member.bic.label('BIC'),
                Member.direct_debit_date.label('Datum'),
            )
            .select_from(Member)
            .join(Allotment)
            .group_by(Member.id)
            .filter(Member.direct_debit == true())
        )


@view_config(route_name='sepa_direct_debit', renderer='json',
             permission='view')
class SEPADirectDebitListView(sw.allotmentclub.browser.base.TableView):

    query_class = SEPADirectDebitQuery



# encoding=utf8
from __future__ import unicode_literals
from datetime import datetime
from io import BytesIO
import openpyxl
import transaction


def setUp():
    from sw.allotmentclub.conftest import import_energy_meters, import_members
    import_members()
    import_energy_meters()
    transaction.commit()


def test_can_download_and_upload_energy_list(browser):
    from sw.allotmentclub import EnergyValue, ElectricMeter
    year = datetime.now().year
    setUp()
    assert [8165, 8411] == [v.value for v in
                            ElectricMeter.get(4).energy_values]
    browser.login()
    browser.open('http://localhost/electricity/export')
    wb = openpyxl.load_workbook(BytesIO(browser.contents))
    sheet = wb.get_active_sheet()
    red = openpyxl.styles.PatternFill(
        start_color='FFFF0000',
        end_color='FFFF0000',
        fill_type='solid')
    for index, row in enumerate(sheet.rows):
        if index in (0, 1):
            continue
        row[-1].value = row[5].value or 0 + 100
        if row[0].value == 4:
            row[-1].fill = red
    to_import = BytesIO()
    wb.save(to_import)
    to_import.seek(0)
    browser._upload(
        'http://localhost/electricity/import',
        ('import.xlsx',
         'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
         to_import))
    assert 16 == (EnergyValue.query()
                  .filter(EnergyValue.year == year).count())
    assert 1 == (EnergyValue.query()
                 .filter(EnergyValue.year == year)
                 .filter(EnergyValue.estimated_value.is_(True)).count())
    assert [8165, 8411, 100] == [v.value for v in
                                 ElectricMeter.get(4).energy_values]



"""Add content for form letters.

Revision ID: 3c0b1fc7eeed
Revises: 12459e1b365d
Create Date: 2015-01-26 14:24:43.103348

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3c0b1fc7eeed'
down_revision = '12459e1b365d'


def upgrade():
    op.add_column('formletter', sa.Column('content', sa.Text(), nullable=True))


def downgrade():
    op.drop_column('formletter', 'content')



"""Add email for users.

Revision ID: 4e4d0ae80e7a
Revises: 56b305a07721
Create Date: 2015-02-27 13:42:09.753241

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4e4d0ae80e7a'
down_revision = '56b305a07721'


def upgrade():
    op.add_column('user',
                  sa.Column('email', sa.String(length=50), nullable=True))
    op.execute("""Update public.user SET email = 'sebastian.wehrmann@me.com'
                  WHERE username = 'sw'""")
    op.execute("""Update public.user SET email = 'ominette@gmx.de'
                  WHERE username = 'ar'""")
    op.execute("""Update public.user SET email = 'smjala@aim.com'
                  WHERE username = 'st'""")
    op.execute("""Update public.user SET email = 'gerd-wmi@t-online.de'
                  WHERE username = 'gm'""")


def downgrade():
    op.drop_column('user', 'email')



"""Add model for Arbeitseinsaetze.

Revision ID: 3ed04edde04e
Revises: 29f31fb4aa
Create Date: 2014-11-10 18:34:35.335596

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3ed04edde04e'
down_revision = '29f31fb4aa'


def upgrade():
    op.create_table(
        'assignment',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('purpose', sa.String(length=254), nullable=True),
        sa.Column('responsible_id', sa.Integer(), nullable=True),
        sa.Column('day', sa.Date(), nullable=True),
        sa.ForeignKeyConstraint(['responsible_id'], [u'member.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    op.create_table(
        'assignmentattendee',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('assignment_id', sa.Integer(), nullable=False),
        sa.Column('member_id', sa.Integer(), nullable=False),
        sa.Column('hours', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['assignment_id'], [u'assignment.id'], ),
        sa.ForeignKeyConstraint(['member_id'], [u'member.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('assignmentattendee')
    op.drop_table('assignment')



"""Make date of assignments a datetime.

Revision ID: 56b305a07721
Revises: 4377c6148413
Create Date: 2015-02-26 13:51:22.336689

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = '56b305a07721'
down_revision = '4377c6148413'


def upgrade():
    op.execute("""ALTER TABLE assignment ALTER COLUMN day TYPE
                  timestamp without time zone;""")
    op.execute("""UPDATE assignment SET day = day + interval '9 hours';""")


def downgrade():
    op.execute("""ALTER TABLE assignment ALTER COLUMN day TYPE date;""")



"""Add bulletin model.

Revision ID: 4377c6148413
Revises: 3f6271158ce3
Create Date: 2015-02-24 10:29:35.004730

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4377c6148413'
down_revision = '3f6271158ce3'


def upgrade():
    op.create_table(
        'bulletin',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('subject', sa.String(), nullable=True),
        sa.Column('content', sa.Text(), nullable=True),
        sa.Column('mimetype', sa.String(length=30), nullable=True),
        sa.Column('size', sa.String(length=20), nullable=True),
        sa.Column('data', sa.LargeBinary(length=10485760), nullable=True),
        sa.Column('date', sa.DateTime(), nullable=False),
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['user_id'], [u'user.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('bulletin')



"""Save calculation of energy values in database.

Revision ID: 186285f37c74
Revises: 2cb8d240d5ac
Create Date: 2015-03-03 10:04:37.845978

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '186285f37c74'
down_revision = '2cb8d240d5ac'


def upgrade():
    op.add_column('energyvalue',
                  sa.Column('advance_pay', sa.Integer(), nullable=True))
    op.add_column('energyvalue',
                  sa.Column('fee', sa.Integer(), nullable=True))
    op.add_column('energyvalue',
                  sa.Column('price', sa.Integer(), nullable=True))
    op.add_column('energyvalue',
                  sa.Column('usage', sa.Integer(), nullable=True))
    op.add_column('energyvalue',
                  sa.Column('whole_price', sa.Integer(), nullable=True))
    op.add_column('energyvalue',
                  sa.Column('discounted', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('energyvalue', 'whole_price')
    op.drop_column('energyvalue', 'usage')
    op.drop_column('energyvalue', 'price')
    op.drop_column('energyvalue', 'fee')
    op.drop_column('energyvalue', 'advance_pay')
    op.drop_column('energyvalue', 'discounted')



"""Add flag if member gets post.

Revision ID: 3f6271158ce3
Revises: dea40636fb3
Create Date: 2015-01-30 12:08:01.977482

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3f6271158ce3'
down_revision = 'dea40636fb3'


def upgrade():
    op.add_column('member',
                  sa.Column('get_post', sa.Boolean(), nullable=True))
    op.execute("""UPDATE member set get_post = true;""")
    op.execute(
        """UPDATE member set get_post = false WHERE id in (190);""")


def downgrade():
    op.drop_column('member', 'get_post')



"""Allow formletters without address field.

Revision ID: 12459e1b365d
Revises: 38a1521607b5
Create Date: 2015-01-26 13:34:52.203956

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '12459e1b365d'
down_revision = '38a1521607b5'


def upgrade():
    op.add_column(
        'formletter', sa.Column('with_address', sa.Boolean(), nullable=True))


def downgrade():
    op.drop_column('formletter', 'with_address')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from .model import Object
from sqlalchemy import Column, String, Boolean, Text, sql
import gocept.loginuser.user
import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm


class User(Object, gocept.loginuser.user.User):

    unrestricted_access = Column(Boolean, nullable=False, default=sql.false())

    log_entries = sqlalchemy.orm.relation(
        'Log', uselist=True, cascade="all,delete", backref='user')

    # Persönliche Daten
    anrede = Column(String(20), default=u'')
    titel = Column(String(20), default=u'')
    vorname = Column(String(50), default=u'')
    nachname = Column(String(50), default=u'')
    handynummer = Column(String(25), default=u'')
    ort = Column(String(50), default=u'')
    position = Column(String(50), default=u'')
    signature = Column(Text())
    email = Column(String(50), default=u'')



"""The allotmentclub portal"""

from setuptools import setup, find_packages
import glob

exec(open("src/sw/allotmentclub/version.py").read())

setup(
    name='sw.allotmentclub',
    version=__version__,  # noqa

    install_requires=[
        'Babel==2.4.0',
        'alembic==0.9.1',
        'apipkg == 1.4',
        'bcrypt==3.1.3',
        'beaker==1.8.1',
        'beautifulsoup4 == 4.6.0',
        'beautifulsoup4==4.6.0',
        'cffi==1.10.0',
        'click==6.7',
        'codecov == 2.0.10',
        'decorator==4.0.11',
        'fints==0.2.0',
        'first==2.0.1',
        'funcsigs==1.0.2',
        'gocept.logging==0.8.1',
        'gocept.loginuser==1.3',
        'html5lib==1.0b10',
        'kontocheck==5.10.0',
        'passlib==1.7.1',
        'lnetatmo==1.4.3',
        'mako==1.0.6',
        'markdown==2.6.8',
        'markupsafe==1.0',
        'mt-940==4.10.0',
        'openpyxl==2.4.8',
        'paste==2.0.3',
        'pastedeploy==1.5.2',
        'pastescript==2.0.2',
        'pillow==4.1.1',
        'pipdeptree==0.10.1',
        'psycopg2==2.7.1',
        'py == 1.5.2',
        'pybars3==0.9.3',
        'pycparser==2.17',
        'pymeta3==0.5.1',
        'pypdf2==1.26.0',
        'pyramid-beaker==0.8',
        'pyramid-exclog==1.0',
        'pyramid-mailer==0.15.1',
        'pyramid-tm==2.0',
        'pyramid-who==0.3',
        'pyramid==1.9.2',
        'python-dateutil==2.6.0',
        'python-editor==1.0.3',
        'pytz==2017.2',
        'radicale==1.1.2',
        'reportlab==3.4.0',
        'repoze.lru==0.6',
        'repoze.sendmail==4.4.1',
        'repoze.who==2.3',
        'requests==2.14.2',
        'risclog.sqlalchemy==3.0',
        'six==1.10.0',
        'sqlalchemy==1.1.10',
        'svglib == 0.8.1',
        'transaction==2.1.2',
        'translationstring==1.3',
        'ua-parser==0.7.3',
        'user-agents==1.1.0',
        'venusian==1.1.0',
        'waitress==1.0.2',
        'webob==1.7.2',
        'xhtml2pdf==0.2b1',
        'xlrd == 1.1.0',
        'xlsxwriter==0.9.6',
        'zope.component==4.3.0',
        'zope.configuration==4.1.0',
        'zope.deprecation==4.2.0',
        'zope.event==4.2.0',
        'zope.exceptions==4.1.0',
        'zope.i18nmessageid==4.1.0',
        'zope.interface==4.4.1',
        'zope.schema==4.4.2',
        'zope.sqlalchemy==0.7.7',
        'zope.testbrowser == 5.2.4',
    ],

    extras_require={
        'test': [
            'coverage == 4.4.2',
            'execnet == 1.5.0',
            'gocept.testdb == 1.3',
            'mock == 2.0.0',
            'pdf-diff3==0.9.3',
            'pep8 == 1.7.1',
            'pytest == 3.3.1',
            'pytest-cache == 1.0',
            'pytest-cov == 2.5.1',
            'pytest-flakes == 2.0.0',
            'pytest-pep8 == 1.0.6',
            'pytest-platform-markers == 0.1.0',
            'pytest-remove-stale-bytecode == 3.0',
            'pytest-rerunfailures == 4.0',
            'pytest-sugar == 0.9.0',
            'termcolor == 1.1.0',
            'unittest2 == 1.1.0',
            'webtest == 2.0.29',
        ],
    },

    entry_points={
        'paste.app_factory': [
            'portal=sw.allotmentclub.browser.app:factory',
        ],
        'console_scripts': [
            'import_members = sw.allotmentclub.scripts:import_members',
            'import_transactions = sw.allotmentclub.scripts:'
            'import_transactions',
            'export_members_vcf = sw.allotmentclub.scripts:'
            'export_members_vcf'
        ],
    },

    author='Sebastian Wehrmann',
    author_email='sebastian@wehrmann.it',
    license='ZPL 2.1',
    url='https://github.com/sweh/sw.allotmentclub',

    keywords='',
    classifiers="""\
        License :: OSI Approved :: Zope Public License
        Programming Language :: Python
        Programming Language :: Python :: 3.5
        Programming Language :: Python :: 3.6
        Programming Language :: Python :: 3 :: Only
        Framework :: Pyramid
        Topic :: Internet :: WWW/HTTP
        Topic :: Internet :: WWW/HTTP :: WSGI :: Application
"""[:-1].split('\n'),
    description=__doc__.strip(),
    namespace_packages=['sw'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    data_files=[('', glob.glob('*.rst'))],
    zip_safe=False,
)



# encoding=utf8
"""Add new lastschrift to members.

Revision ID: 30de960848fa
Revises: bbb62bf367a
Create Date: 2015-11-06 13:54:34.537556

"""

# revision identifiers, used by Alembic.
from __future__ import unicode_literals
from alembic import op

revision = '30de960848fa'
down_revision = u'bbb62bf367a'


def upgrade():
    op.execute("""UPDATE member SET mobile = '0152 28976145',
                                    phone = '034927 21997',
                                    direct_debit = true,
                                    city = 'Kemberg / OT Globig',
                                    bic = 'NOLADE21WBL',
                                    iban = 'DE46805501014623047542',
                                    direct_debit_date = '2015-10-22'
                                    WHERE id = 197;
                """)
    op.execute("""UPDATE member SET firstname = 'André',
                                    direct_debit = true,
                                    bic = 'DRESDEFF88',
                                    iban = 'DE57800800000812092400',
                                    direct_debit_date = '2015-10-27'
                                    WHERE id = 118;
                """)
    op.execute("""UPDATE electricmeter SET allotment_id = 159 WHERE id = 67;
                """)


def downgrade():
    pass



"""re #26: Save message id of sent messages in DB.

Revision ID: ff1bb01af7fa
Revises: 708df911eb9e
Create Date: 2016-11-22 14:40:51.080650

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'ff1bb01af7fa'
down_revision = '708df911eb9e'


def upgrade():
    op.create_table(
        'sentmessageinfo',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('organization_id', sa.Integer(), nullable=False),
        sa.Column('message_id', sa.Integer(), nullable=False),
        sa.Column('msg_id', sa.String(length=100), nullable=True),
        sa.Column('address', sa.String(length=100), nullable=True),
        sa.Column('status', sa.String(length=255), nullable=True),
        sa.ForeignKeyConstraint(['message_id'], [u'message.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('sentmessageinfo')



"""Add revision security flag for users.

Revision ID: 95c83e9e7d18
Revises: 2330cb148c04
Create Date: 2016-05-21 13:57:49.741449

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '95c83e9e7d18'
down_revision = '2330cb148c04'


def upgrade():
    op.add_column(
        'user',
        sa.Column('is_revision', sa.Boolean(),
                  server_default=sa.text(u'false'), nullable=False))


def downgrade():
    op.drop_column('user', 'is_revision')



"""Add mail model.

Revision ID: 43245c2d427e
Revises: 20588bf7b8a7
Create Date: 2015-11-16 09:43:01.683171

"""

# revision identifiers, used by Alembic.
from alembic import op
import sqlalchemy as sa

revision = '43245c2d427e'
down_revision = '20588bf7b8a7'


def upgrade():
    op.create_table(
        'message',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('member_id', sa.Integer(), nullable=True),
        sa.Column('subject', sa.String(length=255), nullable=True),
        sa.Column('body', sa.Text(), nullable=True),
        sa.Column('sent', sa.Date(), nullable=True),
        sa.Column('accounting_year', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(['member_id'], ['member.id'], ),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('message')
